# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""lio.py: Module to manipulate LIO target (using targetcli)."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex

import libsan.host.fc
import libsan.host.linux
from libsan import _print
from libsan.host.cmdline import run

regex_tgtcli_wwpn = "naa.\\S+"


def _tgt_wwn_2_wwn(wwn):
    """On RHEL-6 targetcli stores WWN on WWN format,
    but on RHEL-7 it is something like: naa.200090e2baa397ca
    The arguments are:
    \tNone
    Returns:
    \\tString: WWN as: 20:00:90:e2:ba:a3:97:ca
    """
    # Converting Targetcli wwpn format to common format
    if libsan.host.linux.dist_name() == "RHEL" and libsan.host.linux.dist_ver() < 7:
        return wwn

    wwn_regex = r"naa\."
    wwn = re.sub(wwn_regex, "", wwn)
    # append ":" after every 2nd character
    wwn = re.sub(r"(\S{2})", r"\1:", wwn)
    # remove trail :
    wwn = re.sub(":$", "", wwn)
    return wwn


def _wwn_2_tgt_wwn(wwn):
    """On RHEL-6 targetcli stores WWN on WWN format,
    but on RHEL-7 it is something like: naa.200090e2baa397ca
    The arguments are:
    \tWWN:      20:00:90:e2:ba:a3:97:ca
    Returns:
    \\tString: target WWN foramt as: naa.200090e2baa397ca
    """
    # Converting Targetcli wwpn format to common format
    if libsan.host.linux.dist_name() == "RHEL" and libsan.host.linux.dist_ver() < 7:
        return wwn

    # remove all ':'
    wwn = re.sub(":", "", wwn)
    # append "naa." after every 2nd character
    wwn = "naa." + wwn
    return wwn


def lio_query(show_output=False):
    """Query all information from targetcli using targetcli ls
    The arguments are:
    \tNone
    Returns:
    \tdict: Return a dictionary with targetcli information
    """
    cmd = "targetcli ls"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not run targetcli")
        return None
    lio_data = output.split("\n")

    lio_field_regex = re.compile(r"o-\s(.*?)\s.*\[(.*)\]")
    # supported types for LIO
    lio_supported_types = ("backstores", "iscsi", "loopback", "tcm_fc")
    lio_supported_backstores = ("block", "fileio", "pscsi", "ramdisk", "user:qcow", "user:rbd", "user:zbc")

    lio_dict = {}
    data_type = None

    bs_type_dict = {}

    iscsi_dict = {}
    iscsi_init_iqn = None
    iscsi_tgt_iqn = None
    current_tpg = None
    iscsi_acls_dict = {}
    iscsi_luns = []
    iscsi_portals = []
    iscsi_processing_acls = False
    iscsi_processing_luns = False
    iscsi_processing_portals = False

    tcm_fc_dict = {}
    tcm_fc_wwn = None
    tcm_fc_init_wwn = None
    tcm_fc_acls_dict = {}
    tcm_fc_luns = []
    tcm_fc_processing_acls = False
    tcm_fc_processing_luns = False

    for data in lio_data:
        m = lio_field_regex.search(data)
        if not m:
            # Just ignore entry we can't parse
            # _print("FAIL: (%s) does not match LIO field format" % data)
            continue
        entry = m.group(1)
        entry_details = m.group(2)

        if entry == "/":
            # Skip root
            continue

        # print "INFO: LIO field %s" % entry
        if entry in lio_supported_types:
            data_type = entry
            # bs_type_dict = {}
            lio_dict[data_type] = {}
            continue
        if not data_type:
            _print("FATAL: %s is does not belong to any supported data type" % entry)
            continue
        # print "INFO: %s is subitem of %s" % (entry, data_type)
        # ################# PROCESSING BACKSTORES data type ####################
        if data_type == "backstores":
            if entry in lio_supported_backstores:
                # print "INFO: Processing backstores %s subtiems" % entry
                bs_type = entry
                bs_type_dict[bs_type] = {}
                lio_dict[data_type] = bs_type_dict
                continue
            if entry == "alua" or entry == "default_tg_pt_gp":
                continue
            details_regex = re.compile(r"(.*)\s+\((\S+)\)\s+(\S+)\s+(\S+)")
            details_dict = {}
            m = details_regex.search(entry_details)
            if m:
                details_dict["file_path"] = m.group(1)
                details_dict["lun_size"] = m.group(2)
            details_dict.update(lio_get_backstore_lun_details(bs_type, entry))
            if "wwn" in list(details_dict.keys()):
                details_dict["wwid"] = _lun_wwn2wwid(details_dict["wwn"])
            # print "BRUNO DEBUG backstore %s (%s)" % (entry, entry_details)
            bs_type_dict[bs_type][entry] = details_dict
            lio_dict[data_type] = bs_type_dict

        # ################# PROCESSING iSCSI data type ####################
        if data_type == "iscsi":
            iqn_regex = re.compile(r"iqn\..*")
            if iqn_regex.match(entry) and not iscsi_processing_acls:
                # print "INFO: Processing tcm_fc %s subtiems" % entry
                iscsi_tgt_iqn = entry
                # The target wwn is a dict key
                iscsi_dict[iscsi_tgt_iqn] = {}
                lio_dict[data_type] = iscsi_dict
                continue
            tpg_regex = re.compile(r"(tpg\d+)")
            m = tpg_regex.match(entry)
            if m:
                # print "INFO: Processing tcm_fc %s subtiems" % entry
                current_tpg = m.group(1)
                # The target wwn is a dict key
                iscsi_dict[iscsi_tgt_iqn][current_tpg] = {}
                iscsi_acls_dict = {}
                iscsi_luns = []
                iscsi_portals = []
                lio_dict[data_type] = iscsi_dict
                continue

            if entry == "acls":
                iscsi_dict[iscsi_tgt_iqn][current_tpg]["acls"] = {}
                iscsi_processing_acls = True
                iscsi_processing_luns = False
                iscsi_processing_portals = False
            if entry == "luns":
                iscsi_dict[iscsi_tgt_iqn][current_tpg]["luns"] = {}
                iscsi_processing_acls = False
                iscsi_processing_luns = True
                iscsi_processing_portals = False
                continue
            if entry == "portals":
                iscsi_dict[iscsi_tgt_iqn][current_tpg]["portals"] = {}
                iscsi_processing_acls = False
                iscsi_processing_luns = False
                iscsi_processing_portals = True
                continue
            # ################# PROCESSING ACLS ####################
            # If we are processing ACLs entry
            if iscsi_processing_acls:
                # print "BRUNO ISCSI ACL init (%s)" % entry
                if iqn_regex.match(entry):
                    iscsi_init_iqn = entry
                    iscsi_acls_dict[iscsi_init_iqn] = []
                    iscsi_dict[iscsi_tgt_iqn][current_tpg]["acls"] = iscsi_acls_dict
                    lio_dict[data_type] = iscsi_dict
                    continue
                map_regex = re.compile(r"mapped_(lun.*)$")
                # Check if it is lun mapping information
                m = map_regex.match(entry)
                if m:
                    # print "INFO: found mapped lun: %s" % m.group(1)
                    iscsi_acls_dict[iscsi_init_iqn].append(m.group(1))
                    iscsi_dict[iscsi_tgt_iqn][current_tpg]["acls"] = iscsi_acls_dict
                lio_dict[data_type] = iscsi_dict

            # ################# PROCESSING LUNs ####################
            # If we are processing LUNs entry
            if iscsi_processing_luns:
                iscsi_luns.append(entry)
                iscsi_dict[iscsi_tgt_iqn][current_tpg]["luns"] = iscsi_luns
                lio_dict[data_type] = iscsi_dict
                continue
            # ################# PROCESSING Portlas ####################
            # If we are processing Portals entry
            if iscsi_processing_portals:
                iscsi_portals.append(entry)
                iscsi_dict[iscsi_tgt_iqn][current_tpg]["portals"] = iscsi_portals
                lio_dict[data_type] = iscsi_dict
                continue

        # ################# PROCESSING TCM_FC data type ####################
        if data_type == "tcm_fc":
            # if tcm_fc_processing_luns is true, it is because we reached the end
            # of host wwn and now we are probably processing next host wwn
            # so do not test it on the if below
            tmp_entry = _tgt_wwn_2_wwn(entry)
            if libsan.host.fc.is_wwn(tmp_entry) and not tcm_fc_processing_acls:
                # print "INFO: Processing tcm_fc %s subtiems" % tmp_entry
                tcm_fc_wwn = tmp_entry
                # The target wwn is a dict key
                tcm_fc_dict[tcm_fc_wwn] = {}
                tcm_fc_processing_acls = False
                tcm_fc_processing_luns = False
                lio_dict[data_type] = tcm_fc_dict
                continue

            if entry == "acls":
                tcm_fc_acls_dict = {}
                tcm_fc_dict[tcm_fc_wwn]["acls"] = {}
                tcm_fc_processing_acls = True
                tcm_fc_processing_luns = False
                continue
            if entry == "luns":
                tcm_fc_luns = {}
                tcm_fc_dict[tcm_fc_wwn]["luns"] = {}
                tcm_fc_processing_luns = True
                tcm_fc_processing_acls = False
                continue
            # ################# PROCESSING ACLS ####################
            # If we are processing ACLs entry
            if tcm_fc_processing_acls:
                # It can be initiator, but be using tag instead of wwn
                # TODO: lio_is_fc_tag causes the whole query command to be slow
                # need to find a better way to do it
                tmp_entry = _tgt_wwn_2_wwn(entry)
                if libsan.host.fc.is_wwn(tmp_entry) or lio_is_fc_tag(tcm_fc_wwn, entry):
                    tcm_fc_init_wwn = tmp_entry
                    tcm_fc_acls_dict[tcm_fc_init_wwn] = {}
                    tcm_fc_dict[tcm_fc_wwn]["acls"] = tcm_fc_acls_dict
                    continue
                map_regex = re.compile(r"mapped_(lun.*)$")
                # Check if it is lun mapping information
                m = map_regex.match(entry)
                if m:
                    t_lun_id_regex = re.compile(r"(lun\d+)\s(\S+)/(\S+)")
                    t = t_lun_id_regex.match(entry_details)
                    if t:
                        # print "INFO: found mapped lun: %s" % m.group(1)
                        # print "INFO: entry_details %s" % entry_details
                        tcm_fc_acls_dict[tcm_fc_init_wwn][t.group(1)] = m.group(1)
                        tcm_fc_dict[tcm_fc_wwn]["acls"] = tcm_fc_acls_dict
                        # Update mapping info on backstore session
                        bs_type = t.group(2)
                        lun_name = t.group(3)
                        # print "INFO: tcm_fc acls: bs_type %s lun_name %s is mapped to %s/%s" %
                        # (bs_type, lun_name, tcm_fc_wwn, tcm_fc_init_wwn)
                        details_dict = lio_dict["backstores"][bs_type][lun_name]
                        mapping_dict = {
                            "t_wwpn": tcm_fc_wwn,
                            "h_wwpn": tcm_fc_init_wwn,
                            "t_lun_id": t.group(1),
                            "h_lun_id": m.group(1),
                        }
                        if "mapping" not in list(details_dict.keys()):
                            details_dict["mapping"] = []
                        details_dict["mapping"].append(mapping_dict)

            # ################# PROCESSING LUNS that are added to target wwn ####################
            # If we are processing LUNs entry
            if tcm_fc_processing_luns:
                t_lun_info_regex = re.compile(r"(\S+)/(\S+)\s.*")
                t = t_lun_info_regex.match(entry_details)
                # print "INFO: entry_details tcm_fc luns %s" % entry_details
                if t:
                    # print "INFO: found mapped lun: %s" % m.group(1)
                    # print "INFO: entry_details %s" % entry_details
                    tcm_fc_luns[entry] = {}
                    tcm_fc_luns[entry]["bs_type"] = t.group(1)
                    tcm_fc_luns[entry]["lun_name"] = t.group(2)
                tcm_fc_dict[tcm_fc_wwn]["luns"] = tcm_fc_luns

            lio_dict[data_type] = tcm_fc_dict

    if show_output:
        print(lio_dict)
    return lio_dict


##################################################
# ############### BACKSTORES ######################
##################################################
def lio_create_backstore(bs_type=None, lun_name=None, lun_size=None, device_name=None):
    """Create new backstore device
    The arguments are:
    \tNone
    Returns:
    \tTrue: Device was created
    \tFalse: There was some problem
    """
    created = False
    if bs_type == "block":
        created = _lio_create_backstore_block(lun_name, device_name)

    if bs_type == "fileio":
        created = _lio_create_backstore_fileio(lun_name, lun_size=lun_size)

    if bs_type == "pscsi":
        created = _lio_create_backstore_pscsi(lun_name, device_name)

    if not created:
        _print("FAIL: Could not create lun using (%s) on lio_create_backstore" % bs_type)
        return False

    if lun_name not in list(lio_get_backstores(bs_type).keys()):
        _print("FAIL: It seems %s was created, but it was not" % lun_name)
        return False

    return True


def lio_get_backstores(bs_type=None, lio_dict=None):
    """Return a dict with all backstores. If a backstore type
    is provided return a list of backstore of this type
    The arguments are:
    \tbs_type (Optional): Backstore type
    \tlio_dict (Optional): For optmization, if we already have the lio query no need to do it again
    Returns:
    \tList
    \t\tDict: if there are devices and backstore type was provided
    \tDict
    \t\tDict: All backstore devices
    """
    if not lio_dict:
        lio_dict = lio_query()

    if "backstores" not in list(lio_dict.keys()):
        _print("FAIL: there is not backstore defined on targetcli")
        print(lio_dict)
        return None

    if not bs_type:
        return lio_dict["backstores"]

    if bs_type not in list(lio_dict["backstores"].keys()):
        return None

    return lio_dict["backstores"][bs_type]


def lio_get_backstore_details(bs_type, lun_name, lio_dict=None):
    """Get the size of specifc Backstore device
    Returns:
    \tDic:      Detailed information about this device
    \tNone:     If something went wrong
    """

    bs_dict = lio_get_backstores(bs_type, lio_dict=lio_dict)
    if not bs_dict:
        return None

    if lun_name not in list(bs_dict.keys()):
        _print(f"FAIL: {lun_name} is not defined on {bs_type}")
        lio_dict = lio_query()
        print(lio_dict)
        return None

    return bs_dict[lun_name]


def lio_get_backstore_lun_details(bs_type, lun_name):
    """Get the detailed information about the lun"""
    if not bs_type or not lun_name:
        _print("FAIL: lio_get_backstore_lun_details() - requires bs_type and lun_name")
        return None

    cmd = f"targetcli /backstores/{bs_type}/{lun_name} info"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: Could not get {bs_type} details for {lun_name}")
        return None

    details = output.split("\n")
    supported_details = {}
    supported_details["dev"] = r"^dev: (\S+)"
    supported_details["name"] = r"^name: (\S+)"
    supported_details["size_bytes"] = r"^size: (\S+)"
    supported_details["write_back"] = r"^write_back: (\S+)"
    supported_details["wwn"] = r"^wwn: (\S+)"

    lun_details = {}
    for info in details:
        for sup_detail in supported_details:
            m = re.match(supported_details[sup_detail], info)
            if m:
                lun_details[sup_detail] = m.group(1)

    return lun_details


def _lun_wwn2wwid(wwn):
    """From the LUN WWN is possible to get WWID"""
    wwid = wwn
    wwid = wwid.replace("-", "")
    # Just the first 26 bytes are the wwid
    wwid = wwid[:25]
    wwid = "36001405" + wwid
    return wwid


def lio_delete_backstore(bs_type=None, lun_name=None):
    """Delete backstore device
    The arguments are:
    \tBackstore type
    \tLUN name
    Returns:
    \tTrue: Device was delete
    \tFalse: There was some problem
    """
    deleted = False
    if bs_type == "block":
        deleted = _lio_delete_backstore_block(lun_name)

    if bs_type == "fileio":
        deleted = _lio_delete_backstore_fileio(lun_name)

    if bs_type == "pscsi":
        deleted = _lio_delete_backstore_pscsi(lun_name)

    if not deleted:
        _print("FAIL: Could not delete lun using (%s) on lio_create_backstore" % bs_type)
        return False

    if lun_name in list(lio_get_backstores(bs_type).keys()):
        _print("FAIL: It seems %s was deleted, but it was not" % lun_name)
        return False

    return True


# ## BLOCK ###
def _lio_create_backstore_block(lun_name, device):
    if not lun_name:
        print("FAIL: _lio_create_backstore_block needs lun_name parameter")
        return False
    if not device:
        print("FAIL: _lio_create_backstore_block needs device parameter")
        return False

    cmd = f"targetcli /backstores/block create {lun_name} {device}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not create block %s" % lun_name)
        return False
    return True


def _lio_delete_backstore_block(lun_name):
    cmd = "targetcli /backstores/block delete %s" % lun_name
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete block %s" % lun_name)
        return False

    return True


# ## FILEIO ###
def _lio_create_backstore_fileio(lun_name, file_name=None, lun_size=None):
    if not lun_name:
        _print("_lio_create_backstore_fileio() - requires lun_name parameter")
        return False

    if not file_name:
        # Set default backend file name
        file_name = "%s.img" % lun_name

    # disable spare, to force targetcli to allocate the whole file to avoid problem
    # of running out of disk space and not have enough space to store data
    cmd = f"targetcli /backstores/fileio create {lun_name} {file_name} sparse=false"
    if lun_size:
        cmd = cmd + " %s" % lun_size

    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not create fileio %s" % lun_name)
        return False
    return True


def _lio_delete_backstore_fileio(lun_name):
    file_name = _lio_get_backstore_fileio_file(lun_name)

    cmd = "targetcli /backstores/fileio delete %s" % lun_name
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete fileio %s" % lun_name)
        return False

    if file_name:
        cmd = "rm -f %s" % file_name
        retcode = run(cmd, return_output=False, verbose=True)
        if retcode != 0:
            _print("WARN: could not delete file %s" % file_name)

    return True


def _lio_get_backstore_fileio_file(lun_name):
    """Get the file used by an specific LUN"""
    cmd = "targetcli /backstores/fileio/%s ls" % lun_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not get fileio file %s" % lun_name)
        return None

    path_regex = re.compile(r"\[(.*)\s\(")
    m = path_regex.search(output)
    if m:
        return m.group(1)
    return None


# ## PSCSI ###
def _lio_create_backstore_pscsi(lun_name, device, lun_size=None):
    cmd = f"targetcli /backstores/pscsi create {lun_name} {device}"
    if lun_size:
        cmd = cmd + " %sM" % lun_size

    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not create pscsi %s" % lun_name)
        return False
    return True


def _lio_delete_backstore_pscsi(lun_name):
    cmd = "targetcli /backstores/pscsi delete %s" % lun_name
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete pscsi %s" % lun_name)
        return False

    return True


##################################################
# ################# iSCSI ########################
##################################################
def lio_support_iscsi_target():
    """Check if host supports iSCSI target
    The arguments are:
    \tNone
    Returns:
    \tTrue: Host supports iscsi
    \tFalse: Host does not support iscsi
    """
    lio_dict = lio_query()

    if "iscsi" not in list(lio_dict.keys()):
        # Host does not support iSCSI target
        return False
    return True


# ## iSCSI target ###
def lio_iscsi_create_target(iqn):
    """Add the iqn to iSCSI target
    The arguments are:
    \tiqn:     Target IQN
    Returns:
    \tTrue: If target is added
    \tFalse: If some problem happened
    """
    if not lio_support_iscsi_target():
        _print("FAIL: server does not support iSCSI target")
        return False

    cmd = "targetcli /iscsi/ create %s" % iqn
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not create iSCSI target %s" % iqn)
        return False

    if iqn not in lio_iscsi_get_target():
        _print("FAIL: It seems to have added iSCSI target, but it did not")
        lio_dict = lio_query()
        print(lio_dict["iscsi"])
        return False
    # targetcli by default enable only IPv$ connection
    # we want also IPv6
    if not lio_iscsi_delete_target_portal(iqn, "tpg1", "0.0.0.0"):
        _print("FAIL: could not remove default iSCSI target portal")
        lio_dict = lio_query()
        print(lio_dict["iscsi"])

    if not lio_iscsi_create_target_portal(iqn, "tpg1", "::0"):
        _print("FAIL: could not create IPv6 iSCSI target portal")
        lio_dict = lio_query()
        print(lio_dict["iscsi"])
    return True


def lio_iscsi_get_target():
    """Return a list of all iSCSI targets configured
    The arguments are:
    \tNone
    Returns:
    \tlist: Return a list of IQNs that are configured
    """
    lio_dict = lio_query()

    if not lio_support_iscsi_target():
        # Host does not support iSCSI target
        return None

    return list(lio_dict["iscsi"].keys())


def lio_iscsi_target_set_parameter(tgt_iqn, tpg, group, attr_name, attr_value):
    """Set a parameter to an iSCSI target
    if tgt_iqn is not set, set it globally
    The arguments are:
    \ttgt_iqn       HOST IQN
    \ttpg           Target Portal Group
    \tgroup         eg: attribute, parameter, discovery_auth...
    \tattr_name     Attribute name
    \tattr_value    Attribute value
    Returns:
    \tTrue: If target is attribute is set
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi set {group} {attr_name}={attr_value}"
    if tgt_iqn:
        cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/ set {group} {attr_name}={attr_value}"

    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: Could not set iSCSI target attribute %s" % attr_name)
        return False
    return True


# ## iSCSI ACLS ###
def lio_iscsi_create_acl(tgt_iqn, tpg, init_iqn):
    """Add a initiator IQN to target IQN
    The arguments are:
    \ttgt_iqn:     Host IQN
    \ttpg:         Target Portal Group
    \tinit_iqn:    Initiator IQN
    Returns:
    \tTrue: If init IQN is created
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/acls create {init_iqn} add_mapped_luns=false"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not add iSCSI initiator %s" % init_iqn)
        return False
    return True


def lio_iscsi_delete_acl(tgt_iqn, tpg, init_iqn):
    """Remove a initiator IQN from target IQN
    The arguments are:
    \ttgt_iqn:     Host IQN
    \ttpg:         Target Portal Group
    \tinit_iqn:    Initiator IQN
    Returns:
    \tTrue: If init iqn is removed
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/acls delete {init_iqn}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete iSCSI initiator %s" % init_iqn)
        return False

    return True


# ## iSCSI LUNs ###
def lio_iscsi_add_lun(tgt_iqn, tpg, bs_type, lun_name):
    """Add a LUN to target IQN
    The arguments are:
    \ttgt_iqn:      Host IQN
    \ttpg:          Target Portal Group
    \tbs_type:      Backstore type
    \tlun_name:     Lun Name
    Returns:
    \tTrue: If LUN is added
    \tFalse: If some problem happened
    """
    cmd = "targetcli /iscsi/{}/{}/luns create /backstores/{}/{} add_mapped_luns=false".format(
        tgt_iqn,
        tpg,
        bs_type,
        lun_name,
    )
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not add lun to iSCSI target %s" % tgt_iqn)
        return False

    return True


def lio_iscsi_remove_lun(tgt_iqn, tpg, lun_id):
    """Remove a LUN target IQN
    The arguments are:
    \ttgt_iqn:     Target IQN
    \ttpg:         Target Portal Group
    \tlun_id:      Lun ID
    Returns:
    \tTrue: If LUN is removed
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/luns delete {lun_id}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete LUN from iSCSI target %s" % tgt_iqn)
        return False

    return True


def lio_iscsi_get_luns(tgt_iqn, tpg):
    """Return a list with all LUNs added to an iSCSI target.
    The arguments are:
    \tNone
    Returns:
    \tList: list of luns
    \tNone if something went wrong
    """
    if tgt_iqn not in lio_iscsi_get_target():
        print("FAIL: %s is not defined on targetcli" % tgt_iqn)
        return None

    lio_dict = lio_query()
    if "luns" not in list(lio_dict["iscsi"][tgt_iqn].keys()):
        print("INFO: target %s does not have any LUN\n")
        return None

    return lio_dict["tcm_fc"][tgt_iqn][tpg]["luns"]


# ## iSCSI Portals ###
def lio_iscsi_create_target_portal(tgt_iqn, tpg, portal_ip, portal_port="3260"):
    """Remove a Portal target IQN
    The arguments are:
    \ttgt_iqn:     Target IQN
    \ttpg:         Target Portal Group
    \tportal_ip:   IP of host allowed to connect. (0.0.0.0) any IPv4 address
    \tportal_port  Port to listen for connection, default 3260
    Returns:
    \tTrue: If Portal is created
    \tFalse: If some problem happened
    """
    lio_dict = lio_query()
    if "portals" not in list(lio_dict["iscsi"][tgt_iqn][tpg].keys()):
        print("INFO: target %s does not have support Portal" % tgt_iqn)
        return False

    portal = portal_ip + ":" + portal_port
    if portal in lio_dict["iscsi"][tgt_iqn][tpg]["portals"]:
        print(f"INFO: portal {portal} does already exist on target {tgt_iqn}")
        return True

    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/portals create {portal_ip} {portal_port}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete Portal from iSCSI target %s" % tgt_iqn)
        return False

    return True


def lio_iscsi_delete_target_portal(tgt_iqn, tpg, portal_ip, portal_port="3260"):
    """Remove a Portal target IQN
    The arguments are:
    \ttgt_iqn:     Target IQN
    \ttpg:         Target Portal Group
    \tportal_ip:   IP of host allowed to connect. (0.0.0.0) any IPv4 address
    \tportal_port  Port to listen for connection, default 3260
    Returns:
    \tTrue: If Portal is removed
    \tFalse: If some problem happened
    """
    lio_dict = lio_query()
    if "portals" not in list(lio_dict["iscsi"][tgt_iqn][tpg].keys()):
        print("INFO: target %s does not have support Portal" % tgt_iqn)
        return False

    portal = portal_ip + ":" + portal_port
    if portal not in lio_dict["iscsi"][tgt_iqn][tpg]["portals"]:
        print(f"INFO: portal {portal} does not exist on target {tgt_iqn}")
        return True

    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/portals delete {portal_ip} {portal_port}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete Portal from iSCSI target %s" % tgt_iqn)
        return False
    return True


# ## iSCSI LUNs mapping ###
def lio_iscsi_map_lun(tgt_iqn, tpg, init_iqn, init_lun_id, bs_type, lun_name):
    """Map a LUN to target IQN / Initiator IQN
    The arguments are:
    \ttgt_iqn:      Target IQN
    \ttpg:          Target Portal group
    \tinit_iqn:     Host IQN
    Returns:
    \tTrue: If LUN is mapped
    \tFalse: If some problem happened
    """

    lun_path = f"/backstores/{bs_type}/{lun_name}"
    cmd = f"targetcli /iscsi/{tgt_iqn}/{tpg}/acls/{init_iqn} create {init_lun_id} {lun_path}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print(f"FAIL: Could not map lun to iSCSI target {tgt_iqn}/{init_iqn}")
        return False

    if not lio_iscsi_get_lun_map(tgt_iqn, init_iqn, tpg, init_lun_id):
        _print("FAIL: It seems to have mapped lun %s, but it did not" % init_lun_id)
        return False

    return True


def lio_iscsi_unmap_lun(tgt_iqn, init_iqn, tpg, init_lun_id):
    """Un map LUN from tgt_wwn/init_wwn
    The arguments are:
    \ttgt_iqn:      Target IQN
    \tinit_iqn:     Host IQN
    \ttpg:          Target Portal group
    \tinit_lun_id   LUN ID for the initiator
    Returns:
    \tTrue: If LUN is unmapped
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi/{tgt_iqn}/{init_iqn}/acls/{tpg} delete {init_lun_id}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print(f"FAIL: Could not unmap LUN from target {tgt_iqn}/{init_iqn}")
        return False

    if not lio_iscsi_get_lun_map(tgt_iqn, init_iqn, tpg, init_lun_id):
        _print("FAIL: It seems to have unmapped lun %s, but it did not" % init_lun_id)
        return False
    return True


def lio_iscsi_get_lun_map(tgt_iqn, init_iqn, tpg, tgt_lun_id):
    """Check if a LUN is mapped to target WWN / Initiator port
    The arguments are:
    \ttgt_iqn:      Target IQN
    \tinit_iqn:     Host IQN
    \ttpg:          Target Portal group no.
    \tinit_lun_id   LUN ID for the initiator
    Returns:
    \tTrue: If LUN is mapped
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /iscsi/{tgt_iqn}/tpg{tpg}/acls/{init_iqn} ls | grep {tgt_lun_id}"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: Could not get mapping for lun {tgt_lun_id} on iSCSI target {tgt_iqn}/{init_iqn}")
        return False

    if output == "":
        return False
    return True


def lio_add_iscsi_target(
    tgt_iqn=None, init_iqn=None, bs_type="fileio", lun_name=None, lun_size="1G", device_name=None, tgt_cnt=1, lun_cnt=1
):
    """Create new iSCSI target, create LUNs and do LUN mapping
    The arguments are:
    \ttgt_iqn:          Target IQN, if not specified LIO will create a target IQN
    \tinit_iqn:         Initiator IQN, set LUN map to specific IQN, otherwise any IQN will access the LUN
    \tbs_type:          Backstores storage type, default: 'fileio'
    \tlun_name:         LUN name when creating the target, if not set default lun name will be used
    \tlun_size:         LUN size when using  fileio, default: 1G
    \tdevice_name:      Device name when using block device, for example LV name
    \ttgt_cnt:          Number of targets to create, default: 1
    \tlun_cnt:          Number of LUNs to create, default: 1
    Returns:
    \tTrue: if iSCSI target is created
    \tFalse: If some problem happened
    """

    new_tgt_iqns = []
    iqn_preffix = "iqn.2009-10.com.redhat:storage-"
    # need to create new iSCSI targets, first get the existing targets
    if not tgt_iqn:
        existing_targets = lio_iscsi_get_target()
        iqn_suffix = 0
        # Create new target_iqn names
        while len(new_tgt_iqns) < tgt_cnt:
            tmp_iqn = "%s%d" % (iqn_preffix, iqn_suffix)
            if tmp_iqn not in existing_targets:
                new_tgt_iqns.append(tmp_iqn)
            iqn_suffix += 1
    else:
        new_tgt_iqns = [tgt_iqn]

    for target_iqn in new_tgt_iqns:
        if not lio_iscsi_create_target(target_iqn):
            _print("FAIL: Could not create iSCSI target '%s'" % target_iqn)
            return False

        m = re.match(r"%s(\d+)" % iqn_preffix, target_iqn)
        tgt_name = "tgt%d" % int(m.group(1)) if m else target_iqn.split(":")[1]

        for lun_num in range(1, lun_cnt + 1):
            tgt_lun_name = "%s_lun%d" % (tgt_name, lun_num)
            # If lun name was passed as argument, try to use it
            if lun_name:
                tgt_lun_name = lun_name

            if not lio_create_backstore(
                bs_type=bs_type, lun_name=tgt_lun_name, lun_size=lun_size, device_name=device_name
            ):
                _print("FAIL: Could not create backstore for iSCSI target")
                return False

            tpg = "tpg1"

            if not lio_iscsi_add_lun(target_iqn, tpg, bs_type, tgt_lun_name):
                _print("FAIL: Could not add LUN to iSCSI target")
                return False

            # This is a global setting
            if not lio_iscsi_target_set_parameter(None, None, "discovery_auth", "enable", "0"):
                _print("FAIL: Could not set Attr to iSCSI target")
                return False

            if not lio_iscsi_target_set_parameter(target_iqn, tpg, "attribute", "authentication", "0"):
                _print("FAIL: Could not set Attr to iSCSI target")
                return False

            if not lio_iscsi_target_set_parameter(target_iqn, tpg, "attribute", "generate_node_acls", "1"):
                _print("FAIL: Could not set Attr to iSCSI target")
                return False

            if not lio_iscsi_target_set_parameter(target_iqn, tpg, "attribute", "demo_mode_write_protect", "0"):
                _print("FAIL: Could not set Attr to iSCSI target")
                return False

            lun_id = "0"
            if init_iqn:
                if not lio_iscsi_create_acl(target_iqn, tpg, init_iqn):
                    _print("FAIL: Could not create iSCSI initiator ACL")
                    return False
                if not lio_iscsi_map_lun(target_iqn, tpg, init_iqn, lun_id, bs_type, tgt_lun_name):
                    _print("FAIL: Could not map LUN to iSCSI initiator")
                    return False

    return True


def lio_setup_iscsi_target(
    tgt_iqn=None, init_iqn=None, bs_type="fileio", lun_name=None, lun_size="1G", device_name=None, tgt_cnt=1, lun_cnt=1
):
    """Create a basic iSCSI target
    The arguments are:
    \ttgt_iqn:          Target IQN, if not specified LIO will create a target IQN
    \tinit_iqn:         Initiator IQN, set LUN map to specific IQN, otherwise any IQN will access the LUN
    \tbs_type:          Backstores storage type, default: 'fileio'
    \tlun_name:         LUN name when creating the target, if not set default lun name will be used
    \tlun_size:         LUN size when using  fileio, default: 1G
    \tdevice_name:      Device name when using block device, for example LV name
    \ttgt_cnt:          Number of targets to create, default: 1
    \tlun_cnt:          Number of LUNs to create, default: 1
    Returns:
    \tTrue: if iSCSI target is created
    \tFalse: If some problem happened
    """
    print("INFO: Creating basic iSCSI target...")
    lio_install()
    lio_restart()
    lio_clearconfig()

    ver = lio_version()
    print("INFO: Running targetcli version %s" % ver)

    if not lio_support_iscsi_target():
        _print("FAIL: Server does not support iSCSI target")
        return False

    lio_add_iscsi_target(tgt_iqn, init_iqn, bs_type, lun_name, lun_size, device_name, tgt_cnt, lun_cnt)

    return True


##################################################
# ################# TCM_FC ########################
##################################################
def lio_support_fc_target():
    """Check if host supports FC target
    The arguments are:
    \tNone
    Returns:
    \tTrue: Host supports tcm_fc
    \tFalse: Host does not support tcm_fc
    """
    lio_dict = lio_query()

    if "tcm_fc" not in list(lio_dict.keys()):
        # Host does not support FC target
        return False

    return True


# ## FC target ###
def lio_create_fc_target(wwn):
    """Add the wwn to tcm_fc target
    The arguments are:
    \twwn:     Host wwn
    Returns:
    \tTrue: If target is added
    \tFalse: If some problem happened
    """
    if not wwn:
        _print("FAIL: lio_create_fc_target() - requires wwn parameter")
        return False

    cmd = "targetcli /tcm_fc/ create %s" % wwn
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: lio_create_fc_target() - Could not create FC target %s" % wwn)
        return False

    if wwn not in lio_get_fc_target():
        lio_dict = lio_query()
        run("targetcli ls", return_output=False, verbose=True)
        _print("FAIL: lio_create_fc_target() - It seems to have added FC target, but it did not")
        print(lio_dict["tcm_fc"])
        print(lio_dict)

        return False

    return True


def lio_delete_fc_target(wwn):
    """Delete the wwn to tcm_fc target
    The arguments are:
    \twwn:     Host wwn
    Returns:
    \tTrue: If target is added
    \tFalse: If some problem happened
    """
    if not wwn:
        _print("FAIL: lio_delete_fc_target() - requires wwn parameter")
        return False

    cmd = "targetcli /tcm_fc/ delete %s" % wwn
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: lio_delete_fc_target() - Could not delete FC target %s" % wwn)
        return False

    if wwn in lio_get_fc_target():
        lio_dict = lio_query()
        run("targetcli ls", return_output=False, verbose=True)
        _print("FAIL: lio_delete_fc_target() - It seems to have deleted FC target, but it did not")
        print(lio_dict["tcm_fc"])
        print(lio_dict)

        return False

    return True


def lio_get_fc_target(lio_dict=None):
    """Return a list of all FC targets configured
    The arguments are:
    \tNone
    Returns:
    \tlist: Return a list of wwns that are configured
    """
    if not lio_dict:
        lio_dict = lio_query()

    if not lio_support_fc_target():
        # Host does not support FC target
        return None

    return list(lio_dict["tcm_fc"].keys())


# ## FC ACLS ###
def lio_create_fc_target_acl(tgt_wwn, init_wwn, lio_dict=None):
    """Add a initiator WWN to target WWN port
    The arguments are:
    \ttgt_wwn:     Host WWN
    \tinit_wwn:    Initiator WWN
    Returns:
    \tTrue: If init wwn is created
    \tFalse: If some problem happened
    """
    tgt_acls = lio_get_fc_target_acl(tgt_wwn, lio_dict=lio_dict)
    if tgt_acls and (init_wwn in tgt_acls):
        print(f"INFO: {init_wwn} is already added to target {tgt_wwn}")
        return True

    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/acls create {init_wwn} add_mapped_luns=false"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not add FC initiator %s" % init_wwn)
        return False

    return True


def lio_delete_fc_target_acl(tgt_wwn, init_wwn):
    """Remove a initiator WWN from target WWN port
    The arguments are:
    \ttgt_wwn:     Host WWN
    \tinit_wwn:    Initiator WWN
    Returns:
    \tTrue: If init wwn is removed
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/acls delete {init_wwn}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete FC initiator %s" % init_wwn)
        return False

    return True


def lio_get_fc_target_acl(tgt_wwn, lio_dict=None):
    """Get all acls from an specifc target
    The arguments are:
    \ttgt_wwn:     Host WWN
    Returns:
    \tList: List of initiators
    \tNone: If some problem happened
    """
    if not tgt_wwn:
        _print("FAIL: lio_get_fc_target_acl() - requires tgt_wwpn as argument")
        return None
    if not lio_dict:
        lio_dict = lio_query()
    if not lio_dict["tcm_fc"][tgt_wwn]:
        _print("FAIL: %s does not exist" % tgt_wwn)
        print(lio_dict)
        return None

    if "acls" not in list(lio_dict["tcm_fc"][tgt_wwn].keys()):
        # _print("FAIL: %s does not have acls" % tgt_wwn)
        # print lio_dict
        return None

    return list(lio_dict["tcm_fc"][tgt_wwn]["acls"].keys())


# ## FC LUNs ###
def lio_create_fc_target_lun(tgt_wwn, bs_type, lun_name):
    """Add a LUN to target WWN port
    The arguments are:
    \ttgt_wwn:      Host WWN
    \tbs_type:      Backstore type
    \tlun_name:     Lun Name
    Returns:
    \tTrue: If LUN is created
    \tFalse: If some problem happened
    """
    cmd = "targetcli /tcm_fc/{}/luns create /backstores/{}/{} add_mapped_luns=false".format(
        _wwn_2_tgt_wwn(tgt_wwn),
        bs_type,
        lun_name,
    )
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not add lun to FC target %s" % tgt_wwn)
        return False

    return True


def lio_delete_fc_target_lun(tgt_wwn, lun_id):
    """Remove a initiator WWN from target WWN port
    The arguments are:
    \ttgt_wwn:     Target WWN
    \tlun_id:      Lun ID
    Returns:
    \tTrue: If LUN is removed
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/luns delete {lun_id}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete LUN from target %s" % tgt_wwn)
        return False

    return True


def lio_get_fc_target_luns(tgt_wwn, lio_dict=None):
    """Return a dict with all backstores. If a backstore type
    is provided return a list of backstore of this type
    The arguments are:
    \tNone
    Returns:
    \tList: list of luns
    \tNone if something went wrong
    """
    if not lio_dict:
        lio_dict = lio_query()

    if tgt_wwn not in lio_get_fc_target(lio_dict=lio_dict):
        print("FAIL: %s is not defined on targetcli" % tgt_wwn)
        return None

    if "luns" not in list(lio_dict["tcm_fc"][tgt_wwn].keys()):
        print("INFO: target %s does not have any LUN\n" % tgt_wwn)
        return None

    return lio_dict["tcm_fc"][tgt_wwn]["luns"]


def lio_get_fc_target_lun_id(tgt_wwn, bs_type, lun_name, lio_dict=None):
    """Return the target LUN ID.
    The arguments are:
    \ttgt_wwn:      Target WWN
    \tbs_type:      Backstore Type
    \tlun_name:     LUN name
    Returns:
    \tString: LUN ID. eg: lun0
    \tNone if something went wrong
    """
    if not lio_dict:
        lio_dict = lio_query()

    t_luns_dict = lio_get_fc_target_luns(tgt_wwn, lio_dict)
    if not t_luns_dict:
        return None

    for lun_id in list(t_luns_dict.keys()):
        if t_luns_dict[lun_id]["bs_type"] == bs_type and t_luns_dict[lun_id]["lun_name"] == lun_name:
            return lun_id
    return None


#    cmd = "targetcli /tcm_fc/%s/luns ls | grep %s/%s | awk '{print$2}'" % (_wwn_2_tgt_wwn(tgt_wwn), bs_type, lun_name)
#    retcode, output = run(cmd, return_output=True, verbose=False)
#    if (retcode != 0):
#        _print ("FAIL: Could not get lun %s for FC target %s" % (lun_name, tgt_wwn))
#        return None
#
#    if output == "":
#        return None
#    return output


# ## FC LUNs mapping ###
def lio_fc_lun_map(lun_name, bs_type, tgt_wwn, init_wwn, init_lun_id):
    """Map a LUN to a t_wwpn and h_wwpn"""
    if not lun_name or not bs_type or not tgt_wwn or not init_wwn or not init_lun_id:
        _print("FAIL: lio_fc_lun_map() - requires lun_name, bs_type, tgt_wwn, init_wwn, init_lun_id parameters")
        return False

    _print(f"INFO: Mapping LUN {bs_type}/{lun_name} to {tgt_wwn}/{init_wwn}...")
    lio_dict = lio_query()
    if tgt_wwn not in lio_get_fc_target(lio_dict=lio_dict):
        lio_create_fc_target(tgt_wwn)
        # update lio_dict with new fc target
        lio_dict = lio_query()

    if not lio_get_fc_target_lun_id(tgt_wwn, bs_type, lun_name, lio_dict=lio_dict):
        lio_create_fc_target_lun(tgt_wwn, bs_type, lun_name)
        # update lio_dict with new fc target
        lio_dict = lio_query()

    # Do not pass lio_dict as parameter as we need to query it again to get updatd info
    lun_id = lio_get_fc_target_lun_id(tgt_wwn, bs_type, lun_name, lio_dict=lio_dict)
    if not lun_id:
        _print(f"FAIL: lio_fc_lun_map() - Could not find lun {bs_type}/{lun_name} on target {tgt_wwn}")
        lio_show()
        return False

    if not lio_create_fc_target_acl(tgt_wwn, init_wwn, lio_dict=lio_dict):
        _print("FAIL: Could not create ACL to host %s" % init_wwn)
        lio_show()
        return False

    if not lio_create_fc_target_map_lun(tgt_wwn, init_wwn, init_lun_id, lun_id, lio_dict=lio_dict):
        _print(f"FAIL: Could not map LUN {lun_name} to host {init_wwn}")
        return False

    _print("INFO: LUN %s mapped successfully" % lun_name)
    return True


def lio_create_fc_target_map_lun(tgt_wwn, init_wwn, init_lun_id, tgt_lun_id, lio_dict=None):
    """Map a LUN to target WWN / Initiator port
    The arguments are:
    \ttgt_wwn:      Target WWN
    \tinit_wwn:     Host WWN
    \tinit_lun_id   LUN ID for the initiator
    \ttgt_lun_id:   LUN ID on target
    Returns:
    \tTrue: If LUN is mapped
    \tFalse: If some problem happened
    """

    print("BRUNO lio_create_fc_target_map_lun")
    # print lio_get_fc_target_map_lun(tgt_wwn, init_wwn, tgt_lun_id)

    if lio_get_fc_target_map_lun(tgt_wwn, init_wwn, tgt_lun_id, lio_dict=lio_dict):
        _print(f"INFO: lun {tgt_lun_id} is already mapped to FC target {tgt_wwn}/{init_wwn}")
        return True

    cmd = "targetcli /tcm_fc/{}/acls/{} create {} {}".format(
        _wwn_2_tgt_wwn(tgt_wwn),
        _wwn_2_tgt_wwn(init_wwn),
        init_lun_id,
        tgt_lun_id,
    )
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print(f"FAIL: Could not map lun to FC target {tgt_wwn}/{init_wwn}")
        return False

    if not lio_get_fc_target_map_lun(tgt_wwn, init_wwn, tgt_lun_id):
        _print("FAIL: It seems to have mapped lun %s, but it did not" % tgt_lun_id)
        return False

    return True


def lio_fc_target_get_mapped_luns(tgt_wwn, init_wwn, lio_dict=None):
    """Get LUN mapping from tgt_wwn/init_wwn
    The arguments are:
    \ttgt_wwn:      Target WWN
    \tinit_wwn:     Host WWN
    Returns:
    \tDict:         Dictionary with tgt_lunid : init_lun_id
    \tNone:         No mapping was found
    """

    if not lio_dict:
        lio_dict = lio_query()

    if init_wwn not in lio_get_fc_target_acl(tgt_wwn, lio_dict=lio_dict):
        return None

    mapped_luns_dict = lio_dict["tcm_fc"][tgt_wwn]["acls"][init_wwn]
    if not mapped_luns_dict:
        return None
    return mapped_luns_dict


def lio_fc_target_unmap_lun(tgt_wwn, init_wwn, init_lun_id):
    """Un map LUN from tgt_wwn/init_wwn
    The arguments are:
    \ttgt_wwn:      Target WWN
    \tinit_wwn:     Host WWN
    \tinit_lun_id   LUN ID for the initiator
    Returns:
    \tTrue: If LUN is unmapped
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/acls/{_wwn_2_tgt_wwn(init_wwn)} delete {init_lun_id}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print(f"FAIL: Could not unmap LUN from target {tgt_wwn}/{init_wwn}")
        return False

    if lio_get_fc_target_map_lun(tgt_wwn, init_wwn, init_lun_id):
        _print("FAIL: It seems to have unmapped lun %s, but it did not" % init_lun_id)
        return False
    return True


def lio_get_fc_target_map_lun(tgt_wwn, init_wwn, tgt_lun_id, lio_dict=None):
    """Get initator LUN Id if a LUN is mapped to target WWN / Initiator port
    The arguments are:
    \ttgt_wwn:      Target WWN
    \tinit_wwn:     Host WWN
    \ttgt_lun_id:   LUN ID on target
    Returns:
    \tinit_lun_id: If LUN is mapped
    \tNone: If LUN is not mapped
    """

    # If lio dict is given as parameter we do not need to query lio output again
    if not lio_dict:
        lio_dict = lio_query()

    if init_wwn not in lio_get_fc_target_acl(tgt_wwn, lio_dict=lio_dict):
        return None

    # print "DEBUG lio_get_fc_target_map_lunt t: %s" % tgt_wwn
    # print "DEBUG lio_get_fc_target_map_lunt i: %s" % init_wwn
    # print "DEBUG lio_get_fc_target_map_lunt tgt_id: %s" % tgt_lun_id
    init_acls_dict = lio_dict["tcm_fc"][tgt_wwn]["acls"][init_wwn]
    if tgt_lun_id not in list(init_acls_dict.keys()):
        # _print ("FAIL: Could not get mapping for lun %s on FC target %s/%s" % (tgt_lun_id, tgt_wwn, init_wwn))
        return None

    return init_acls_dict[tgt_lun_id]


def lio_get_fc_target_lun_mapping(bs_type, lun_name, lio_dict=None):
    """Get the mapping information for an specifc LUN
    The arguments are:
    \tbs_type:      Backstore Type
    \tlun_name:     LUN name
    Returns:
    \tList:         A list with a dictionary for each mapping found
    \tNone:         If no mapping was found
    """
    if not lio_dict:
        lio_dict = lio_query()

    bs_details_dict = lio_get_backstore_details(bs_type, lun_name, lio_dict=lio_dict)
    if not bs_details_dict:
        # It does not exist
        return None

    if "mapping" not in list(bs_details_dict.keys()):
        return None

    return bs_details_dict["mapping"]


# ## FC tag ###
def lio_tag_fc_initiator(tgt_wwn, init_wwn, tag):
    """Create a tag for initiator wwn
    The arguments are:
    \ttgt_wwn:     Host wwn
    \tinit_wwn:    Initiator wwn
    \ttag:          tag for the initiator wwn
    Returns:
    \tTrue: If tag is created
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/acls tag {_wwn_2_tgt_wwn(init_wwn)} {tag}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not tag FC initiator %s" % init_wwn)
        return False

    return True


def lio_untag_fc_initiator(tgt_wwn, tag):
    """Remove tag from initiator wwn
    The arguments are:
    \ttgt_wwn:     Host wwn
    \ttag:          tag for the initiator wwn
    Returns:
    \tTrue: If tag is created
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{_wwn_2_tgt_wwn(tgt_wwn)}/acls untag {tag}"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not untag FC tag %s" % tag)
        return False

    return True


def lio_is_fc_tag(tgt_wwn, tag):
    """Check if a tag is a FC initiator tag
    The arguments are:
    \twwn:     Host wwn
    Returns:
    \tTrue: If target is added
    \tFalse: If some problem happened
    """
    cmd = f"targetcli /tcm_fc/{tgt_wwn}/acls/{tag}"
    retcode = run(cmd, return_output=False, verbose=False)
    if retcode != 0:
        return False
    return True


##################################################
# ################ LIO General ####################
##################################################
def lio_install():
    """Install targetcli tool
    The arguments are:
    \tNone
    Returns:
    \tTrue: If targetcli is installed correctly
    \tFalse: If some problem happened
    """
    ver = libsan.host.linux.dist_ver()
    if not ver:
        return False

    targetcli_pack = "targetcli"
    if libsan.host.linux.dist_name() == "RHEL" and ver < 7:
        targetcli_pack = "fcoe-target-utils"

    if not libsan.host.linux.install_package(targetcli_pack):
        _print("FAIL: Could not install %s" % targetcli_pack)
        return False

    return True


def lio_get_service_name():
    """Get service name that targetcli uses
    The arguments are:
    \tNone
    Returns:
    \tString: Name of the service
    \tNone: If some problem happened
    """
    targetcli_service = "target"
    ver = libsan.host.linux.dist_ver()
    if not ver:
        return None

    if libsan.host.linux.dist_name() == "RHEL" and ver < 7:
        targetcli_service = "fcoe-target"

    return targetcli_service


def lio_restart():
    """Restart LIO service
    The arguments are:
    \tNone
    Returns:
    \tTrue: Service started
    \tFalse: If some problem happened
    """
    targetcli_service = lio_get_service_name()
    if not targetcli_service:
        _print("FAIL: lio_restart() - Could not get LIO service name")
        return False

    if not libsan.host.linux.service_restart(targetcli_service):
        _print("FAIL: Could not restart LIO service")
        return False
    # sleep 5s to avoid service to not be restarted
    # target.service start request repeated too quickly, refusing to start.
    libsan.host.linux.sleep(5)
    return True


def lio_show():
    """List LIO configuration
    The arguments are:
    \tNone
    Returns:
    \tTrue: If listed config
    \tFalse: If some problem happened
    """
    cmd = "targetcli ls"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not show LIO config")
        return False
    return True


def lio_saveconfig():
    """Save LIO configuration
    The arguments are:
    \tNone
    Returns:
    \tTrue: If config is saved
    \tFalse: If some problem happened
    """
    cmd = "targetcli saveconfig"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not save LIO config")
        return False
    return True


def lio_clearconfig():
    """Clear LIO configuration
    The arguments are:
    \tNone
    Returns:
    \tTrue: If config is deleted
    \tFalse: If some problem happened
    """
    _print("INFO: Cleaning up LIO configuration")

    if not libsan.host.linux.is_installed("targetcli"):
        return True

    fileio_dict = lio_get_backstores("fileio")
    if fileio_dict:
        # Delete all files before cleaning configuration
        for lun in list(fileio_dict.keys()):
            lio_delete_backstore(bs_type="fileio", lun_name=lun)

    cmd = "targetcli clearconfig true"
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete LIO config")
        return False

    return True


def lio_version():
    """Get targetcli version
    The arguments are:
    \tNone
    Returns:
    \tString: TargetCli version
    \tNone: If some problem happened
    """
    cmd = "targetcli version"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not get targetcli version")
        return None

    version_regex = re.compile(".* version (.*)$")
    m = version_regex.search(output)
    if m:
        return m.group(1)
    _print("FAIL: Could not parse targetcli version output (%s)" % output)
    return None


def lio_clean_up_targets(lio_dict=None):
    """Removing backstore might leave targets with empty mapping
    They should be removed
    """
    lio_dict = lio_query(lio_dict)
    fc_targets = lio_get_fc_target(lio_dict=lio_dict)
    if not fc_targets:
        # nothing to clean up
        return True
    # First we checked for initiators without mapping
    need_to_query_lio = False
    success = True
    for tgt in fc_targets:
        initiators = lio_get_fc_target_acl(tgt, lio_dict=lio_dict)
        if initiators:
            for init in initiators:
                if not lio_fc_target_get_mapped_luns(tgt, init, lio_dict=lio_dict):
                    _print(f"DEBUG: Should remove initiator {init} from tgt {tgt}")
                    if not lio_delete_fc_target_acl(tgt, init):
                        success = False
                    need_to_query_lio = True

    # Check again for targets without any initiator
    if need_to_query_lio:
        lio_dict = lio_query()
    fc_targets = lio_get_fc_target(lio_dict=lio_dict)
    for tgt in fc_targets:
        initiators = lio_get_fc_target_acl(tgt, lio_dict=lio_dict)
        if not initiators:
            _print("DEBUG: Should remove target %s" % tgt)
            if not lio_delete_fc_target(tgt):
                success = False

    return success


class TargetCLI:
    def __init__(self, path="", disable_check=False):
        self.disable_check = disable_check
        self.path = path
        if libsan.host.linux.dist_ver() < 7:
            _print("FATAL: TargetCLI is not supported on RHEL < 7.")

        if not libsan.host.linux.install_package("targetcli", check=False):
            _print("FATAL: Could not install targetcli package")

    @staticmethod
    def remove_nones(kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    @staticmethod
    def _extract_args(kwargs, keys=None):
        keys = keys or ["return_output", "verbosity", "path"]
        arguments = {}
        for key in keys:
            if key not in kwargs:
                continue
            arguments[key] = kwargs.pop(key)
        return arguments, kwargs

    def _run(self, cmd, verbosity=True, return_output=False, path=None):
        # Constructs the command to run and runs it

        if path is not None:
            self.path = path

        cmd = "targetcli cd" if cmd == "cd" and self.path is None else "targetcli " + self.path + " " + cmd

        if return_output:
            ret, data = run(cmd, verbose=verbosity, return_output=True)
            if ret != 0:
                _print("WARN: Running command: '%s' failed. Return with output." % cmd)
            return ret, data

        ret = run(cmd, verbose=verbosity)
        if ret != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
        return ret

    def ls(self, depth="", **kwargs):
        return self._run("ls %s" % depth, **kwargs)

    def cd(self, **kwargs):
        return self._run("cd", **kwargs)

    def pwd(self, **kwargs):
        return self._run("pwd", **kwargs)

    def create(self, **kwargs):
        keys = None
        cmd = "create "
        arguments, kwargs = self._extract_args(kwargs)
        # the following ensures the ordering is correct in correct paths
        # True means it is required, False it is optional
        if "backstores/block" in self.path:
            keys = {"name": True, "dev": True, "readonly": False, "wwn": False}
        elif "backstores/fileio" in self.path:
            keys = {"name": True, "file_or_dev": True, "size": True, "write_back": False, "sparse": False, "wwn": False}
        elif "backstores/pscsi" in self.path:
            keys = {"name": True, "dev": True}
        elif "backstores/ramdisk" in self.path:
            keys = {"name": True, "size": True, "nullio": False, "wwn": False}
        elif "backstores/user:qcow" in self.path:  # noqa: SIM114
            keys = {
                "name": True,
                "size": True,
                "cfgstring": True,
                "wwn": False,
                "hw_max_sectors": False,
                "control": False,
            }
        elif "backstores/user:rbd" in self.path:  # noqa: SIM114
            keys = {
                "name": True,
                "size": True,
                "cfgstring": True,
                "wwn": False,
                "hw_max_sectors": False,
                "control": False,
            }
        elif "backstores/user:zbc" in self.path:
            keys = {
                "name": True,
                "size": True,
                "cfgstring": True,
                "wwn": False,
                "hw_max_sectors": False,
                "control": False,
            }
        elif self.path.startswith("/iscsi/"):
            if "iqn" in self.path:
                if "acls" in self.path:
                    keys = {"wwn": True, "add_mapped_luns": False}
                elif "luns" in self.path:
                    keys = {"storage_object": True, "lun": False, "add_mapped_luns": False}
                elif "portals" in self.path:
                    keys = {"ip_address": False, "ip_port": False}
                else:
                    keys = {"tag": False}
            else:
                keys = {"wwn": False}
        elif self.path.startswith("/loopback"):
            if "naa" in self.path:
                if "luns" in self.path:
                    keys = {"storage_object": True, "lun": False, "add_mapped_luns": False}
            else:
                keys = {"wwn": False}
        else:
            keys = {"wwn": False}
        for key in keys:
            try:
                cmd += f"{key}={kwargs[key]} "
            except KeyError:
                if not self.disable_check and keys[key]:
                    _print(f"FAIL: Create on path '{self.path}' requires argument {key}.")
                    return 1
        return self._run(cmd, **arguments)

    def delete(self, **kwargs):
        keys = None
        cmd = "delete "
        arguments, kwargs = self._extract_args(kwargs)
        # the following ensures the ordering is correct in correct paths
        # True means it is required, False it is optional
        if "backstores/block" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/fileio" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/pscsi" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/ramdisk" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/user:qcow" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/user:rbd" in self.path:  # noqa: SIM114
            keys = {"name": True}
        elif "backstores/user:zbc" in self.path:
            keys = {"name": True}
        elif self.path.startswith("/iscsi/"):
            if "iqn" in self.path:
                if "acls" in self.path:
                    keys = {"wwn": True}
                elif "luns" in self.path:
                    keys = {"lun": True}
                elif "portals" in self.path:
                    keys = {"ip_address": True, "ip_port": True}
                else:
                    keys = {"tag": True}
            else:
                keys = {"wwn": True}
        elif self.path.startswith("/loopback"):
            if "naa" in self.path:
                if "luns" in self.path:
                    keys = {"lun": True}
            else:
                keys = {"wwn": False}
        else:
            keys = {"wwn": True}

        for key in keys:
            try:
                cmd += f"{key}={kwargs[key]} "
            except KeyError:
                if not self.disable_check and keys[key]:
                    _print(f"FAIL: Delete on path '{self.path}' requires argument {key}.")
                    return 1

        return self._run(cmd, **arguments)

    def help(self, topic="", **kwargs):
        return self._run("help %s" % topic, **kwargs)

    def saveconfig(self, savefile=None, **kwargs):
        cmd = "saveconfig"
        if savefile:
            cmd += " %s" % savefile
        return self._run(cmd, **kwargs)

    def restoreconfig(self, savefile="/etc/target/saveconfig.json", clearexisting=False, **kwargs):
        return self._run(f"restoreconfig {savefile} {clearexisting}", **kwargs)

    def clearconfig(self, confirm=True, **kwargs):
        return self._run("clearconfig %s" % confirm, **kwargs)

    def sessions(self, action="", sid="", **kwargs):
        return self._run(f"sessions {action} {sid}", **kwargs)

    def exit(self, **kwargs):
        return self._run("exit", **kwargs)

    def get(self, group="", **kwargs):
        arguments, kwargs = self._extract_args(kwargs)
        cmd = "get {} {}".format(group, " ".join(kwargs.keys()))
        return self._run(cmd, **arguments)

    def set(self, group="", **kwargs):
        arguments, kwargs = self._extract_args(kwargs)
        params = [
            f"{kwarg}='{kwargs[kwarg]}'" if kwargs[kwarg] != "" else f"{kwarg}='{kwargs[kwarg]}'" for kwarg in kwargs
        ]
        cmd = "set {} {}".format(group, " ".join(params))
        return self._run(cmd, **arguments)

    def info(self, **kwargs):
        return self._run("info", **kwargs)

    def version(self, **kwargs):
        return self._run("version", **kwargs)

    def status(self, **kwargs):
        return self._run("status", **kwargs)

    def refresh(self, **kwargs):
        return self._run("refresh", **kwargs)

    def disable(self, **kwargs):
        return self._run("disable", **kwargs)

    def enable(self, **kwargs):
        return self._run("enable", **kwargs)

    def bookmarks(self, **kwargs):
        # How to use this?
        return self._run("bookmarks", **kwargs)

    def enable_iser(self, boolean="", **kwargs):
        return self._run("enable_iser %s" % boolean, **kwargs)

    def enable_offload(self, boolean="", **kwargs):
        return self._run("enable_offload %s" % boolean, **kwargs)

    def tag(self, wwn_or_tag="", new_tag="", **kwargs):
        return self._run(f"tag {wwn_or_tag} {new_tag}", **kwargs)

    def untag(self, wwn_or_tag="", **kwargs):
        return self._run("untag %s" % wwn_or_tag, **kwargs)



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import io
import re
import sys
import unittest

from libsan.host.cmdline import run


class TestCmdline(unittest.TestCase):
    def test_run_verbose(self):
        test_msg = "run test message"
        # write to stderr, the message should be shown in the test
        test_cmd = f"echo {test_msg} >&2"
        running_regex = rf"INFO: \[\d+-\d+-\d+ \d+:\d+:\d+\] Running: '{test_cmd}'..."
        cmd_output_regex = f"^{test_msg}$"

        new_callable = io.StringIO()
        sys.stdout = new_callable
        # By default is verbose true
        assert run(test_cmd) == 0
        sys.stdout = sys.__stdout__
        assert re.search(running_regex, new_callable.getvalue().split("\n")[0])
        cmd_output_regex = f"^{test_msg}$"
        assert re.search(cmd_output_regex, new_callable.getvalue().split("\n")[1])
        sys.stdout = sys.__stdout__

        # not verbose
        new_callable = io.StringIO()
        sys.stdout = new_callable
        assert run(test_cmd, verbose=False) == 0
        sys.stdout = sys.__stdout__
        assert new_callable.getvalue() == ""
        sys.stdout = sys.__stdout__

    def test_run_force_flush(self):
        test_msg = "run test message"
        test_cmd = f"echo {test_msg}"
        running_regex = rf"INFO: \[\d+-\d+-\d+ \d+:\d+:\d+\] Running: '{test_cmd}'..."
        cmd_output_regex = f"^{test_msg}$"

        new_callable = io.StringIO()
        sys.stdout = new_callable
        assert run(test_cmd, force_flush=True) == 0
        sys.stdout = sys.__stdout__
        assert re.search(running_regex, new_callable.getvalue().split("\n")[0])
        cmd_output_regex = f"^{test_msg}$"
        assert re.search(cmd_output_regex, new_callable.getvalue().split("\n")[1])
        sys.stdout = sys.__stdout__

    def test_run_return_output(self):
        test_msg = "run test message"
        test_cmd = f"echo {test_msg}"
        running_regex = rf"INFO: \[\d+-\d+-\d+ \d+:\d+:\d+\] Running: '{test_cmd}'..."
        cmd_output_regex = f"^{test_msg}$"

        new_callable = io.StringIO()
        sys.stdout = new_callable
        assert run(test_cmd, return_output=True) == (0, test_msg)
        sys.stdout = sys.__stdout__
        assert re.search(running_regex, new_callable.getvalue().split("\n")[0])
        cmd_output_regex = f"^{test_msg}$"
        assert re.search(cmd_output_regex, new_callable.getvalue().split("\n")[1])
        sys.stdout = sys.__stdout__

    def test_run_fail_return_output(self):
        failure_msg = "ls: cannot access 'invalid_file': No such file or directory"
        test_cmd = "ls invalid_file"
        running_regex = rf"INFO: \[\d+-\d+-\d+ \d+:\d+:\d+\] Running: '{test_cmd}'..."
        cmd_output_regex = f"^{failure_msg}$"

        new_callable = io.StringIO()
        sys.stdout = new_callable
        assert run(test_cmd, return_output=True) == (2, failure_msg)
        sys.stdout = sys.__stdout__
        assert re.search(running_regex, new_callable.getvalue().split("\n")[0])
        cmd_output_regex = f"^{failure_msg}$"
        assert re.search(cmd_output_regex, new_callable.getvalue().split("\n")[1])
        sys.stdout = sys.__stdout__

    def test_run_timeout(self):
        test_cmd = "sleep 9"
        expected_retcode = 124
        assert run(test_cmd, return_output=True, timeout=1) == (expected_retcode, "")
        assert run(test_cmd, return_output=True, timeout=1, force_flush=True) == (expected_retcode, "")



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from unittest.mock import call, patch

from libsan.host import md

mdstat_output = """Personalities : [raid0]
md0 : active raid0 sdb[1] sda[0]
      2093056 blocks super 1.2 512k chunks

unused devices: <none>
"""

mdadm_output = """
/dev/md0:
           Version : 1.2
     Creation Time : Fri Nov 18 04:15:54 2022
        Raid Level : raid0
        Array Size : 2093056 (2044.00 MiB 2143.29 MB)
      Raid Devices : 2
     Total Devices : 2
       Persistence : Superblock is persistent

       Update Time : Fri Nov 18 04:15:54 2022
             State : clean
    Active Devices : 2
   Working Devices : 2
    Failed Devices : 0
     Spare Devices : 0

            Layout : -unknown-
        Chunk Size : 512K

Consistency Policy : none

              Name : 0
              UUID : 05f704a3:70531edf:80b2414a:0de2efad
            Events : 0

    Number   Major   Minor   RaidDevice State
       0       7        0        0      active sync   /dev/sda
       1       7        1        1      inactive      /dev/sdb1
"""

md0_info = {
    "version": "1.2",
    "creation_time": "Fri Nov 18 04:15:54 2022",
    "raid_level": "raid0",
    "array_size": "2093056 (2044.00 MiB 2143.29 MB)",
    "raid_devices": "2",
    "total_devices": "2",
    "persistence": "Superblock is persistent",
    "update_time": "Fri Nov 18 04:15:54 2022",
    "state": "clean",
    "active_devices": "2",
    "working_devices": "2",
    "failed_devices": "0",
    "spare_devices": "0",
    "layout": "-unknown-",
    "chunk_size": "512K",
    "name": "0",
    "uuid": "05f704a3:70531edf:80b2414a:0de2efad",
    "events": "0",
    "storage_devices": {
        "/dev/sda": {"number": "0", "major": "7", "minor": "0", "raid_device": "0", "state": "active sync"},
        "/dev/sdb1": {"number": "1", "major": "7", "minor": "1", "raid_device": "1", "state": "inactive"},
    },
}


class TestMD(unittest.TestCase):
    @patch("libsan.host.md.run")
    def test_md_query(self, run_func):
        run_func.return_value = (0, mdstat_output)

        assert md.md_query() == ["md0"]

        assert run_func.call_count == 1
        run_calls = [call("cat /proc/mdstat", return_output=True, verbose=False)]
        run_func.assert_has_calls(run_calls)

    @patch("libsan.host.md._mdadm_query")
    @patch("libsan.host.md.md_query")
    def test_md_get_info(self, md_query_func, mdadm_query_func):
        md_query_func.return_value = ["md0"]
        mdadm_query_func.return_value = mdadm_output

        assert md.md_get_info("md0", verbose=True) == md0_info

    @patch("libsan.host.md.md_get_info")
    def test_md_get_storage_dev(self, md_get_info_func):
        md_get_info_func.return_value = md0_info

        storage_devs = md.md_get_storage_dev("md0")
        assert "/dev/sda" in storage_devs
        assert "/dev/sdb1" in storage_devs

    @patch("libsan.host.md.run")
    def test_md_stop(self, run_func):
        run_func.return_value = (0, "")
        assert md.md_stop("md0")
        run_calls = [call("mdadm --stop /dev/md0", return_output=True, verbose=False)]
        run_func.assert_has_calls(run_calls)

    @patch("libsan.host.md.run")
    def test_md_clean(self, run_func):
        run_func.return_value = (0, "")
        assert md.md_clean("/dev/sda")
        run_calls = [call("mdadm --zero-superblock /dev/sda", return_output=True, verbose=False)]
        run_func.assert_has_calls(run_calls)

    @patch("libsan.host.md.run")
    @patch("libsan.host.md.md_stop")
    @patch("libsan.host.md.md_get_storage_dev")
    def test_md_remove(self, get_sto_func, md_stop_func, run_func):
        run_func.return_value = (0, "")
        md_stop_func.return_value = True
        get_sto_func.return_value = ["/dev/sda1", "/dev/sdb1"]
        assert md.md_remove("md0", clean=True)
        run_calls = [
            call("mdadm --remove /dev/md0", return_output=True, verbose=False),
            call("mdadm --zero-superblock /dev/sda1", return_output=True, verbose=False),
            call("mdadm --zero-superblock /dev/sdb1", return_output=True, verbose=False),
        ]
        run_func.assert_has_calls(run_calls)



# Copyright (C) 2018 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""cli_tools.py: Module to provide tools of wrapping command line tools for further usage."""

__author__ = "Jakub Krysl"
__copyright__ = "Copyright (c) 2018 Red Hat, Inc. All rights reserved."

from libsan import _print


class WrongCommandExceptionError(Exception):
    def __init__(self, cmd):
        self.command = cmd
        super().__init__()

    def __str__(self):
        return repr(WrongCommandExceptionError.__name__ + ", caused by " + repr(self.command))


class WrongArgumentExceptionError(Exception):
    def __init__(self, arg, cmd=None, args=None):
        self.argument = arg
        self.command = cmd
        self.arguments = args
        super().__init__()

    def __str__(self):
        return repr(WrongArgumentExceptionError.__name__ + ", caused by " + repr(self.argument))


class FailedCheckExceptionError(Exception):
    def __init__(self, arg=None):
        self.argument = arg
        super().__init__()

    def __str__(self):
        message = repr(FailedCheckExceptionError.__name__)
        if self.argument:
            message += ", caused by " + repr(self.argument)
        return message


class Wrapper:
    def __init__(self, commands, arguments, disable_check):
        self.commands = commands
        self.arguments = arguments
        self.disable_check = disable_check

    def _add_command(self, cmd):
        # Checks if given command is provided by CLI and returns its correct syntax
        if cmd in self.commands:
            return self.commands[cmd]
        raise WrongCommandExceptionError(cmd)

    def _get_arg(self, name):
        if not self.disable_check:
            if name in self.arguments:
                return self.arguments[name][1]
            raise WrongArgumentExceptionError(name)
        return self.arguments[name][1]

    def _get_cmd(self, name):
        if self.disable_check:
            return self.commands["all"]
        if name in self.arguments:
            return self.arguments[name][0]
        raise WrongCommandExceptionError(name)

    @staticmethod
    def _get_value(string, command, return_type=str):
        _value = string.split(command)[1].split()[0]
        try:
            value = return_type(_value)
        except ValueError as e:
            _print("WARN: Got ValueError: %s." % e)
            return None
        return value

    def _get_possible_arguments(self, command=None):
        # Returns possible arguments for said command if specified
        args = []
        if command:
            for key in list(self.arguments.keys()):
                if list(self.commands.keys())[list(self.commands.values()).index(command)] in self._get_cmd(key):
                    args.append(key)
        else:
            args = list(self.arguments.keys())
        return args

    @staticmethod
    def _add_value(value, command, argument):
        if argument[-1:] in ["=", "&"]:
            if argument[-1:] == "&":
                argument = argument[:-1] + " "
            if isinstance(value, list):
                # allows to use repeatable arguments as a list of values
                for val in value:
                    command += argument + "'" + str(val) + "'"
            else:
                command += argument + "'" + str(value) + "'"
        else:
            command += argument
        return command

    def _check_allowed_argument(self, arg, command):
        if arg not in self.arguments and not self.disable_check:
            raise WrongArgumentExceptionError(arg)
        cmd = command.split()[0]
        args = self._get_possible_arguments(cmd)
        if arg not in args:
            raise WrongArgumentExceptionError(arg, cmd, args)

    def _add_argument(self, arg, value, command):
        # Checks if given argument is allowed for given command and adds it to cmd string
        self._check_allowed_argument(arg, command)
        command = self._add_value(value, command, self._get_arg(arg))
        return command

    def _add_arguments(self, cmd, **kwargs):
        command = cmd
        for kwarg in kwargs:
            # skip adding this argument if the value is False
            if kwargs[kwarg] is False:
                continue
            command = self._add_argument(kwarg, kwargs[kwarg], command)
        return command



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""dm.py: Module to manipulate Device mapper devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex

from libsan import _print
from libsan.host.cmdline import run


def dm_show_status(dm_device=None):
    """Show dmsetup status.
    The arguments are:
    \tdm_device:           Device name (optional)
    Returns:
    \tTrue
    or
    \tFalse
    """
    cmd = "dmsetup status"
    if dm_device:
        cmd += " %s" % dm_device

    retcode, _ = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not show dmsetup status")
        return False
    return True


def dm_query_status(dm_device=None):
    """Query dmsetup status and return a dictonary with table output for each device.
    The arguments are:
    \tdm_device:           Device name (optional)
    Returns:
    \tdict: Return a dictionary with status info for each device
    """
    cmd = "dmsetup status"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not list status")
        return None
    devices = output.split("\n")

    status_basic_format_regex = r"(.*):\s+(\d+)\s+(\d+)\s+(\S+)"
    # Thin formats are available on: https://www.kernel.org/doc/Documentation/device-mapper/thin-provisioning.txt
    status_thin_format_regex = status_basic_format_regex + r"\s+(\d+)\s+(\d+|-)"
    status_thin_pool_format_regex = status_basic_format_regex + r"\s+(\d+)\s+(\d+)\/(\d+)\s+(\d+)\/(\d+)(.*)"
    status_multipath_format_regex = (
        status_basic_format_regex + r"\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(.*)"
    )

    dev_status_dict = {}
    for device in devices:
        m = re.match(status_basic_format_regex, device)
        if not m:
            _print("FAIL: dm_query_status() - (%s) does not match dmsetup status output format" % device)
            continue
        dm_type = m.group(4)
        dm_status_format_dict = {}
        if dm_type == "linear":
            dm_status_format_dict = {"logical_start_sector": m.group(2), "num_sec": m.group(3), "target_type": dm_type}
        elif dm_type == "thin":
            t = re.match(status_thin_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_status_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "nr_mapped_sec": t.group(5),
                "highest_mapped_sec": t.group(6),
            }
        elif dm_type == "thin-pool":
            t = re.match(status_thin_pool_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_status_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "trans_id": t.group(5),
                "used_metadata": t.group(6),
                "total_metadata": t.group(7),
                "used_data_block": t.group(8),
                "total_data_block": t.group(9),
                "metadata": t.group(10),
            }
        elif dm_type == "multipath":
            t = re.match(status_multipath_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_status_format_dict = {
                "logical_start_sector": m.group(2),
                "num_sec": m.group(3),
                "target_type": dm_type,
                "nr_status_feature": t.group(5),
                "all_io_queued": t.group(6),
                "nr_att_init_grp": t.group(7),
                "nr_hw_handlers": t.group(8),
                "nr_path_grps": t.group(9),
                "active_path_grp": t.group(10),
            }
            remaining_data = t.group(11)

            status_multipath_path_args_format_regex = r"\s+(\S+)"
            status_multipath_path_format_regex = r"\s+(\S+)\s+(\S+)\s+(\S+)"
            status_multipath_grp_header_format_regex = r"\s?(\S)\s+(\d+)\s+(\d+)\s+(\d+)"

            nr_groups = int(dm_status_format_dict["nr_path_grps"])
            for group_nr in range(1, int(nr_groups) + 1):
                grp_match = re.match(status_multipath_grp_header_format_regex, remaining_data)
                if not grp_match:
                    _print("FAIL: dm_query_status() - Could not parse group multipath data")
                    return None
                if "path_grp" not in list(dm_status_format_dict.keys()):
                    dm_status_format_dict["path_grp"] = {}
                dm_status_format_dict["path_grp"][group_nr] = {}
                path_grp = dm_status_format_dict["path_grp"][group_nr]
                path_grp["state"] = grp_match.group(1)
                path_grp["nr_grp_status_arg"] = grp_match.group(2)
                path_grp["nr_paths"] = grp_match.group(3)
                path_grp["nr_path_status_arg"] = grp_match.group(4)
                nr_paths = int(path_grp["nr_paths"])  # FIXME: Expected type 'Union[int, slice]', got 'str' instead
                remaining_data = re.sub(status_multipath_grp_header_format_regex, "", remaining_data, 1)
                for p_nr in range(1, int(nr_paths) + 1):
                    path_match = re.match(status_multipath_path_format_regex, remaining_data)
                    if not path_match:
                        _print("FAIL: dm_query_status() - Could not parse path multipath data")
                        return None
                    if "path" not in list(path_grp.keys()):
                        path_grp["path"] = {}
                    path_grp["path"][p_nr] = {}  # FIXME: Expected type 'Union[int, slice]', got 'str' instead
                    path = path_grp["path"][p_nr]  # FIXME: Expected type 'Union[int, slice]', got 'str' instead
                    path["device"] = path_match.group(1)
                    path["state"] = path_match.group(2)
                    path["failure_cnt"] = path_match.group(3)
                    remaining_data = re.sub(status_multipath_path_format_regex, "", remaining_data, 1)
                    for _ in range(1, int(grp_match.group(4)) + 1):
                        path_arg_match = re.match(status_multipath_path_args_format_regex, remaining_data)
                        if not path_arg_match:
                            _print("FAIL: dm_query_status() - Could not parse path args multipath data")
                            return None
                        if "path_status_args" not in list(path.keys()):
                            path["path_status_args"] = ""
                        path["path_status_args"] += "%s " % path_arg_match.group(1)
                        remaining_data = re.sub(status_multipath_path_args_format_regex, "", remaining_data, 1)
                    # remove trailing space from the end of string
                    if "path_status_args" in list(path.keys()):
                        path["path_status_args"] = path["path_status_args"].strip()

        else:
            _print(f"FAIL: dm_query_status() - ({device}) can't parse status for device type {dm_type}")

        dev_status_dict[m.group(1)] = dm_status_format_dict
    if dm_device:
        if dm_device in list(dev_status_dict.keys()):
            return dev_status_dict[dm_device]
        return None
    return dev_status_dict


def dm_show_table(dm_device=None):
    """Show dmsetup table.
    The arguments are:
    \tdm_device:           Device name (optional)
    Returns:
    \tTrue
    or
    \tFalse
    """
    cmd = "dmsetup table"
    if dm_device:
        cmd += " %s" % dm_device

    retcode, _ = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not show dmsetup table")
        return False
    return True


def dm_query_table(dm_device=None):
    """Query dmsetup table and return a dictonary with table output for each device.
    The arguments are:
    \tdm_device:           Device name (optional)
    Returns:
    \tdict: Return a dictionary with table info for each device
    """
    cmd = "dmsetup table"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not list table")
        return None
    devices = output.split("\n")

    # Basic format information is found on man dmsetup under "TABLE FORMAT" section
    # format of dmsetup table: logical_start_sector num_sectors target_type target_args
    table_basic_format_regex = r"(.*):\s+(\S+)\s+(\S+)\s+(\S+)"
    # format for linear type
    table_linear_format_regex = table_basic_format_regex + r"\s+(\S+)\s+(\S+)$"
    # vgtest-thin1: 0 2097152 thin 253:5 1
    table_thin_format_regex = table_basic_format_regex + r"\s+(\S+)\s+(\S+)$"
    # vgtest-test_pool-tpool: 0 1638400 thin-pool 253:3 253:4 512 0 1 error_if_no_space
    table_thin_pool_format_regex = table_basic_format_regex + r"\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)$"
    # delayed: 0 20971520 delay 7:0 0 100
    table_delay_format_regex = table_basic_format_regex + r"\s+(\S+)\s+(\S+)\s+(\S+)$"

    dev_table_dict = {}
    for device in devices:
        m = re.match(table_basic_format_regex, device)
        if not m:
            _print("FAIL: (%s) does not match dmsetup table output format" % device)
            continue
        dm_type = m.group(4)
        dm_table_format_dict = {}
        if dm_type == "linear":
            t = re.match(table_linear_format_regex, device)
            if not t:
                _print("FAIL: (%s) does not match dmsetup a valid linear table output format" % device)
                continue
            dm_table_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "dev": t.group(5),
                "start_sec": t.group(6),
            }
        elif dm_type == "thin":
            t = re.match(table_thin_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_table_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "pool_dev": t.group(5),
                "dev_num": t.group(6),
            }
        elif dm_type == "thin-pool":
            t = re.match(table_thin_pool_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_table_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "metadata_dev": t.group(5),
                "data_dev": t.group(6),
                "data_block_size": t.group(7),
                "low_water_mark": t.group(8),
                "target_args": t.group(9),
            }
        elif dm_type == "delay":
            t = re.match(table_delay_format_regex, device)
            if not t:
                _print(f"FAIL: ({device}) does not match dmsetup a valid {dm_type} table output format")
                continue
            dm_table_format_dict = {
                "logical_start_sector": t.group(2),
                "num_sec": t.group(3),
                "target_type": dm_type,
                "dev": t.group(5),
                "start_sec": t.group(6),
                "delay": t.group(7),
            }
        elif dm_type == "multipath":
            # multipath table format see: http://christophe.varoqui.free.fr/refbook.html
            device_params = device.split(" ")
            dm_table_format_dict = {}
            dm_table_format_dict["logical_start_sector"] = m.group(2)
            dm_table_format_dict["num_sec"] = m.group(3)
            dm_table_format_dict["target_type"] = dm_type
            dm_table_format_dict["nr_status_feature"] = int(device_params[4])
            # set index to next entry after nr_status_feature
            param_index = 4 + 1
            if dm_table_format_dict["nr_status_feature"] > 0:
                dm_table_format_dict["features"] = []
                for _ in range(dm_table_format_dict["nr_status_feature"]):
                    dm_table_format_dict["features"].append(device_params[param_index])
                    param_index += 1

            dm_table_format_dict["nr_hw_handlers"] = int(device_params[param_index])
            param_index += 1
            if dm_table_format_dict["nr_hw_handlers"] > 0:
                dm_table_format_dict["hw_handlers"] = []
                for _ in range(dm_table_format_dict["nr_hw_handlers"]):
                    dm_table_format_dict["hw_handlers"].append(device_params[param_index])
                    param_index += 1

            dm_table_format_dict["nr_path_grps"] = int(device_params[param_index])
            param_index += 1
            dm_table_format_dict["try_next_path_grp"] = device_params[param_index]
            param_index += 1

            remaining_data = ""
            for _ in range(len(device_params) - param_index):
                remaining_data += "%s " % device_params[param_index]
                param_index += 1
            remaining_data = remaining_data.strip()

            table_multipath_grp_header_format_regex = r"\s?(\S+)\s+(\d+)\s+(\d+)\s+(\d+)"
            table_multipath_path_header_format_regex = r"\s?(\S+)"
            table_multipath_path_args_format_regex = r"\s?(\S+)"
            nr_groups = dm_table_format_dict["nr_path_grps"]
            for group_nr in range(1, int(nr_groups) + 1):
                grp_match = re.match(table_multipath_grp_header_format_regex, remaining_data)
                if not grp_match:
                    _print("FAIL: dm_query_table() - Could not parse group multipath data")
                    return None
                if "path_grp" not in list(dm_table_format_dict.keys()):
                    dm_table_format_dict["path_grp"] = {}
                dm_table_format_dict["path_grp"][group_nr] = {}
                path_grp = dm_table_format_dict["path_grp"][group_nr]
                path_grp["path_selector"] = grp_match.group(1)
                path_grp["nr_selector_args"] = grp_match.group(2)
                path_grp["nr_paths_grp"] = grp_match.group(3)
                path_grp["nr_path_args"] = grp_match.group(4)
                # remove processed data
                remaining_data = re.sub(table_multipath_grp_header_format_regex, "", remaining_data, 1)
                for path_nr in range(1, int(path_grp["nr_paths_grp"]) + 1):
                    path_match = re.match(table_multipath_path_header_format_regex, remaining_data)
                    if not path_match:
                        _print("FAIL: dm_query_table() - Could not parse path multipath data")
                        return None
                    if "path" not in list(dm_table_format_dict.keys()):
                        dm_table_format_dict["path"] = {}
                    dm_table_format_dict["path"][path_nr] = {}
                    path = dm_table_format_dict["path"][path_nr]
                    path["device"] = path_match.group(1)

                    remaining_data = re.sub(table_multipath_path_header_format_regex, "", remaining_data, 1)
                    for _ in range(1, int(path_grp["nr_path_args"]) + 1):
                        arg_match = re.match(table_multipath_path_args_format_regex, remaining_data)
                        if not arg_match:
                            _print("FAIL: dm_query_table() - Could not parse path arg multipath data")
                            return None
                        if "path_status_args" not in list(path.keys()):
                            path["path_status_args"] = ""
                        path["path_status_args"] += "%s " % arg_match.group(1)
                        remaining_data = re.sub(table_multipath_path_args_format_regex, "", remaining_data, 1)
                    if "path_status_args" in list(path.keys()):
                        # remove trailing space
                        path["path_status_args"] = path["path_status_args"].strip()

        else:
            _print(f"FAIL: dm_query_table() - ({device}) can't parse table for device type '{dm_type}'")
        dev_table_dict[m.group(1)] = dm_table_format_dict

    if dm_device:
        if dm_device in list(dev_table_dict.keys()):
            return dev_table_dict[dm_device]
        return None
    return dev_table_dict


def dm_get_table_device(dm_name):
    """Get table information for specific device.
    The table info is not parsed
    """
    if not dm_name:
        return None
    cmd = "dmsetup table %s" % dm_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not query table for %s" % dm_name)
        return None

    return output


def dm_set_target_type(dm_name, target_type):
    """Change the target type of an specific DM device
    it is necessary to suspend the device before loading new table
    The arguments are:
    \tDevice name: eg. VG-lv_home
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse in case of failure
    """
    table_format_regex = r"(\S+)\s+(\S+)\s+(\S+)\s(.*)$"
    table_info = dm_get_table_device(dm_name)
    if not table_info:
        # Could not get table for device name
        return False

    # We need to suspend the device first
    if not dm_suspend_dev(dm_name):
        return False

    m = re.match(table_format_regex, table_info)
    if not m:
        _print("FAIL: dm_set_target_type() - (%s) does not match dmsetup table output format" % table_info)
        return False
    # load the table with new target type
    new_table = f'"{m.group(1)} {m.group(2)} {target_type} {m.group(4)}"'

    cmd = f"dmsetup load {dm_name} --table {new_table}"
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: dm_set_target_type() - could not load table on %s" % dm_name)
        dm_resume_dev(dm_name)
        return False

    if not dm_resume_dev(dm_name):
        return False

    return True


def dm_suspend_dev(dm_name):
    """Executes dmsetup suspend to suspend an specific DM device
    The arguments are:
    \tDevice name: eg. VG-lv_home
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse in case of failure
    """
    cmd = "dmsetup suspend %s" % dm_name
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: could not suspend %s" % dm_name)
        return False

    return True


def dm_resume_dev(dm_name):
    """Executes dmsetup resume to suspend an specific DM device
    The arguments are:
    \tDevice name: eg. VG-lv_home
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse in case of failure
    """
    cmd = "dmsetup resume %s" % dm_name
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: could not resume %s" % dm_name)
        return False

    return True


def dm_message_dev(dm_name, message):
    """Executes dmsetup message to send a message to an specific DM device
    The arguments are:
    \tDevice name: eg. mpatha
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse in case of failure
    """
    cmd = f"dmsetup message {dm_name} {message}"
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: could not send message to %s" % dm_name)
        return False

    return True


def dm_remove(dm_name):
    """Remove the the specified device
    The arguments are:
    \tDevice name: eg. Device mapped device name
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse in case of failure
    """
    devs = dm_query_table()
    if dm_name not in list(devs.keys()):
        # device name does not exist
        return True

    cmd = "dmsetup remove %s" % dm_name
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: could not remove %s" % dm_name)
        return False
    return True



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import io
import sys
import unittest

import libsan
import pytest


class TestLibsan(unittest.TestCase):
    def test_print(self):
        # simple print
        new_callable = io.StringIO()
        sys.stdout = new_callable
        libsan._print("test _print")
        sys.stdout = sys.__stdout__
        assert new_callable.getvalue() == "test _print\n"
        sys.stdout = sys.__stdout__

        # DEBUG print
        new_callable = io.StringIO()
        sys.stdout = new_callable
        libsan._print("DEBUG: test _print")
        sys.stdout = sys.__stdout__
        assert new_callable.getvalue() == "DEBUG:(%s) test _print\n" % __name__
        sys.stdout = sys.__stdout__

        # FATAL print
        new_callable = io.StringIO()
        sys.stdout = new_callable
        with pytest.raises(RuntimeError, match=r"FATAL:\(tests.libsan_test\) test _print"):
            libsan._print("FATAL: test _print")
        sys.stdout = sys.__stdout__
        assert new_callable.getvalue() == "FATAL:(%s) test _print\n" % __name__
        sys.stdout = sys.__stdout__



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""fc.py: Module to manipulate FC devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os
import os.path
import re  # regex

import libsan.host.linux
import libsan.host.scsi
from libsan import _print
from libsan.host.cmdline import run

host_path = "/sys/class/fc_host/"

remote_port_path = "/sys/class/fc_remote_ports/"

regex_target_id = re.compile(r"target(\d+):(\d+):(\d+)")
regex_target = re.compile(r"(\d+):(\d+):(\d+)")
wwn_regex = re.compile(r"(?:[0-9a-f]{2}:){7}[0-9a-f]{2}")


def is_wwn(wwn):
    """Checks if the entry is on valid WWN format.
    example: 10:00:5c:b9:01:c1:ec:71
    The arguments are:
    \ta WWN
    Returns:
    \tTrue if it is a valid WWN
    \tFalse if the entry is not valid.
    """
    m = wwn_regex.match(wwn)
    if m:
        return True
    return False


def standardize_wwpn(wwpn):
    """Usage
        standardize_wwpn(wwpn)
    Purpose
        Convert all possiable format wwpn into stand type:
            (?:[0-9a-f]{2}:){7}[0-9a-f]{2} #like: 10:00:00:00:c9:95:2f:de
        Return STRING or ARRAY base on context.
    Parameter
        wwpn           # any format wwpn, like "500A0981894B8DC5"
    Returns
        wwpn.
    """
    if not wwpn:
        return None
    wwpn = wwpn.lower()
    if is_wwn(wwpn):
        return wwpn
    wwpn = re.sub("0x", "", wwpn)
    # remove all : characters from the entry, later on they will be added in correct order
    wwpn = wwpn.replace(":", "")
    # append ":" every 2nd character
    wwpn = ":".join(wwpn[i : i + 2] for i in range(0, len(wwpn), 2))
    if is_wwn(wwpn):
        return wwpn
    return None


def query_all_fc_disks(host_id=None):
    """Return SCSI disk info all all FC/FCoE devices."""
    hosts = get_fc_hosts()
    if not hosts:
        # there is not FC/FCoE session
        return None

    if host_id:
        hosts = [host_id]

    scsi_disks = libsan.host.scsi.query_all_scsi_disks()
    fc_disks = {}
    for scsi_id in list(scsi_disks.keys()):
        scsi_disk = scsi_disks[scsi_id]
        if scsi_disk["host_id"] in hosts:
            fc_disks[scsi_id] = scsi_disk
    return fc_disks


def scsi_wwid_of_fc_disks():
    """Return a list of all WWIDs of FC/FCoE disks."""
    disks_info = query_all_fc_disks()
    if not disks_info:
        return None
    wwids = []
    for info in list(disks_info.values()):
        if "wwid" not in list(info.keys()):
            continue
        if info["wwid"] and info["wwid"] not in wwids:
            wwids.append(info["wwid"])
    return wwids


def is_fc_boot():
    """Check if it boots from FC/FCoE device."""
    fc_wwids = scsi_wwid_of_fc_disks()
    if not fc_wwids:
        return False

    boot_dev = libsan.host.linux.get_boot_device()
    if not boot_dev:
        _print("FAIL: is_fc_boot() - Could not determine boot device")
        return False

    boot_wwid = libsan.host.linux.get_device_wwid(boot_dev)
    if not boot_dev:
        _print("WARN: is_fc_boot() - Could not determine boot WWID for %s" % boot_dev)
        return False

    if boot_wwid in fc_wwids:
        return True

    return False


# Return an array with all fc_hosts numbers
def get_fc_hosts():
    cmd = "ls " + host_path
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    # remove 'host' prefix
    output = re.sub("host", "", output)
    host_array = output.split()
    return host_array


def get_fc_host_wwpn(host):
    """Return the WWPN of specific host."""
    host_port_path = "/sys/class/fc_host/host%s/port_name" % host
    cmd = "cat " + host_port_path
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    wwpn = standardize_wwpn(output)
    return wwpn


def fc_host_id_of_wwpn(wwpn):
    """Given a WWPN, return its host id."""
    if not wwpn:
        _print("FAIL: fc_host_id_of_wwpn() - requires wwpn parameter")
        return None
    fc_hosts = get_fc_hosts()
    for fc_host in fc_hosts:
        if get_fc_host_wwpn(fc_host) == wwpn:
            return fc_host

    return None


def h_wwpn_of_host(host=None):
    """h_wwpn_of_host
    Usage:
        h_wwpn_of_host(host)
        h_wwpn_of_host()
    Purpose:
        Return a list with the WWPN of specific host
        If no host is given return all WwPNs
    Parameter
        host (options).

    Return:
    ------
        List:   WWPNs
        None:   Could not find any WWPN.
    """
    wwpns = None
    hosts = [host] if host else get_fc_hosts()
    if not hosts:
        return None
    for host in hosts:
        wwpn = get_fc_host_wwpn(host)
        if wwpn:
            if not wwpns:
                wwpns = []
            wwpns.append(wwpn)
    return wwpns


def get_rports():
    rports = os.listdir(remote_port_path)
    if not rports:
        return None
    return rports


def get_fc_host_remote_ports(host):
    cmd = f"ls {remote_port_path} | grep rport-{host}"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    r_ports_array = output.split()
    return r_ports_array


def wwpn_of_rport(r_port):
    cmd = f"cat {remote_port_path}/{r_port}/port_name"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    wwpn = standardize_wwpn(output)
    return wwpn


def rport_of_h_wwpn_t_wwpn(h_wwpn, t_wwpn):
    """Return the remote port of given wwpn
    Return:
        r_port       #like rport-7:0-0
            or
        None.
    """
    h_wwpn = standardize_wwpn(h_wwpn)
    if not h_wwpn:
        return None
    t_wwpn = standardize_wwpn(t_wwpn)
    if not t_wwpn:
        return None

    host_id = fc_host_id_of_wwpn(h_wwpn)
    if not host_id:
        return None

    # Get all remote ports from this host
    host_rports = get_fc_host_remote_ports(host_id)
    if not host_rports:
        return None

    for rport in host_rports:
        r_wwpn = wwpn_of_rport(rport)
        if not r_wwpn:
            # _print("FAIL: rport_of_h_wwpn_t_wwpn() - Could not get wwpn of rport %s" % rport)
            continue
        if t_wwpn == r_wwpn:
            return rport
    return None


def fc_target_id_of_wwpn(wwpn):
    """ """
    if not wwpn:
        _print("FAIL: fc_target_id_of_wwpn() - requires wwpn parameter")
        return None
    t_ids = None

    # Get target id of all ports
    rport_id_2_target_id_dict = rport_id_2_target_id()
    if not rport_id_2_target_id_dict:
        return None

    fc_hosts = get_fc_hosts()
    if not fc_hosts:
        return None

    for fc_host in fc_hosts:
        r_ports = get_fc_host_remote_ports(fc_host)
        if not r_ports:
            continue
        for r_port in r_ports:
            if wwpn_of_rport(r_port) == wwpn:
                if not t_ids:
                    t_ids = []
                if r_port in list(rport_id_2_target_id_dict.keys()):
                    t_ids.append(rport_id_2_target_id_dict[r_port])

    return t_ids


def t_wwpn_of_host(host=None):
    """t_wwpn_of
    Usage:
        t_wwpn_of(host)
        t_wwpn_of()
    Purpose:
        Return a list with the WWPN of targets of specific host
        If no host is given return all WwPNs
    Parameter
        host (options).

    Return:
    ------
        List:   WWPNs
        None:   Could not find any WWPN.
    """
    wwpns = None
    hosts = [host] if host else get_fc_hosts()
    if not hosts:
        return None

    # Host wwpns are shwon on remote port, we should not return them as target ports
    h_wwpns = h_wwpn_of_host()

    for host in hosts:
        r_ports = get_fc_host_remote_ports(host)
        if r_ports:
            for r_port in r_ports:
                wwpn = wwpn_of_rport(r_port)
                if not wwpn:
                    # _print("FAIL: Could not find wwpn for r_port %s" % r_port)
                    continue
                if wwpn in h_wwpns:
                    # We actually found our own WWPN, skip it
                    continue
                if not wwpns:
                    wwpns = []
                if wwpn not in wwpns:
                    wwpns.append(wwpn)
    return wwpns


def fc_target_id_of_htwwpn(t_wwpn=None, h_wwpn=None):
    """Usage
        fc_target_id_of_htwwpn(h_wwpn=>h_wwpn, t_wwpn=>t_wwpn)
    Purpose
        Query out the target_id came from h_wwpn and t_wwpn. As FC has FSPF,
        protocol, we will only got ONE target_id for 1 combination.
    Parameter
        h_wwpn         # Host HBA WWPN
        t_wwpn         # Storage Array front point WWPN
    Returns
        target_id      # like "0:1:0"
            or
        None           # error.
    """
    if not t_wwpn or not h_wwpn:
        _print("FAIL: fc_target_id_of_htwwpn() - requires t_wwpn and h_wwpn parameters")
        return None

    scsi_host_id = fc_host_id_of_wwpn(h_wwpn)
    target_ids = fc_target_id_of_wwpn(t_wwpn)
    if not target_ids:
        _print("FAIL: fc_target_id_of_htwwpn(): No FC target assigned to " + "target WWPN %s" % t_wwpn)
        return None

    for t_id in target_ids:
        m = regex_target.match(t_id)
        if m and scsi_host_id == m.group(1):
            return t_id
    return None


def scsi_disk_of_htwwpn_wwid(h_wwpn, t_wwpn, wwid):
    """Get the scsi disk name connected to specific host and target port and has
    given wwid.
    """
    if not h_wwpn or not t_wwpn or not wwid:
        print("FAIL: scsi_disk_of_htwwpn_wwid() - requires h_wwpn, t_wwpn and wwid as parameters")
        return None

    target_id = fc_target_id_of_htwwpn(t_wwpn=t_wwpn, h_wwpn=h_wwpn)
    if not target_id:
        return None

    scsi_disk_ids = libsan.host.scsi.scsi_ids_of_wwid(wwid)
    if not scsi_disk_ids:
        return False

    for scsi_disk_id in scsi_disk_ids:
        if re.match(r"%s:\d+$" % target_id, scsi_disk_id):
            disk_name = libsan.host.scsi.get_scsi_disk_name(scsi_disk_id)
            if not disk_name:
                _print("FAIL: Could not get SCSI disk name of: %s" % scsi_disk_id)
                return False
            return disk_name
    return None


def get_fc_host_rport_targets(host, r_port):
    cmd = f"ls /sys/bus/scsi/devices/host{host}/{r_port} | grep target"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    output = re.sub("target", "", output)
    r_port_targets_array = output.split()
    return r_port_targets_array


def get_fc_host_rport_target_devices(host, r_port, target):
    if not host or not r_port or not target:
        _print("FAIL: get_fc_host_rport_target_devices. Usage: host, r_port, target")
        return None

    cmd = f'ls /sys/bus/scsi/devices/host{host}/{r_port}/target{target} | grep "{target}"'
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    output = re.sub("target", "", output)
    devices_array = output.split()
    return devices_array


def rport_id_2_target_id():
    """Usage
        rport_id_2_target_id()
    Purpose
        For lpfc card, their rport_id is different from target_id.
        This function will return this reference:
            rport_id_2_target_id = {
                rport_id : target_id
            }
        Only the rport which provide FC target will return.

    Returns
    -------
        rport_id_2_target_id
            or
        None.
    """
    if not os.path.isdir(remote_port_path):
        return None

    rport_id_2_target_id_dict = None
    rports = get_rports()
    if not rports:
        return None
    for rport in rports:
        # search for targets connected to this rport
        rport_target_sys_root_path = f"{remote_port_path}/{rport}/device"
        entries = os.listdir(rport_target_sys_root_path)

        for entry in entries:
            m = regex_target_id.match(entry)
            if m:
                if not rport_id_2_target_id_dict:
                    rport_id_2_target_id_dict = {}
                rport_id_2_target_id_dict[rport] = f"{m.group(1)}:{m.group(2)}:{m.group(3)}"

    return rport_id_2_target_id_dict


def fc_host_transport_type(scsi_host_id):
    """ """
    if not scsi_host_id:
        _print("FAIL: fc_host_transport_type - requires scsi_host_id")
        return None

    host_model2transport = {
        "QLE2462": "FC",
        "QLE2772": "FC",
        "QLE8262": "FCoE",
        "QLE8362": "FCoE",
        "CN1000Q": "FCoE",
        "QLogic-1020": "FCoE",
        "554FLR-SFP+": "FCoE",
        "Intel 82599": "FCoE",
    }
    host_info = libsan.host.scsi.query_scsi_host_info(scsi_host_id)
    # Check first if we can find what transport by HBA model
    model_keys = ["model", "model_name"]
    for key in model_keys:
        if key in list(host_info.keys()) and host_info[key] in list(host_model2transport.keys()):
            return host_model2transport[host_info[key]]

    # Second option try to parse model description
    if "model_description" in list(host_info.keys()):
        fc_regex1 = re.compile("Fibre Channel")
        fc_regex2 = re.compile(" FC ")
        fcoe_regex = re.compile("FCoE")
        if fc_regex1.search(host_info["model_description"]):
            return "FC"
        if fc_regex2.search(host_info["model_description"]):
            return "FC"
        if fcoe_regex.search(host_info["model_description"]):
            return "FCoE"

    if "model_desc" in list(host_info.keys()):
        fc_regex1 = re.compile("Fibre Channel")
        fc_regex2 = re.compile(" FC ")
        fcoe_regex = re.compile("FCoE")
        if fc_regex1.search(host_info["model_desc"]):
            return "FC"
        if fc_regex2.search(host_info["model_desc"]):
            return "FC"
        if fcoe_regex.search(host_info["model_desc"]):
            return "FCoE"

    if "modeldesc" in list(host_info.keys()):
        fc_regex = re.compile("Fibre Channel")
        fcoe_regex = re.compile("FCoE")
        if fc_regex.search(host_info["modeldesc"]):
            return "FC"
        if fcoe_regex.search(host_info["modeldesc"]):
            return "FCoE"

    if "protocol" in list(host_info.keys()):
        if host_info["protocol"] == "fc":
            return "FC"
        if host_info["protocol"] == "fcoe":
            return "FCoE"

    if "symbolic_name" in list(host_info.keys()):
        fc_regex = re.compile("Fibre Channel")
        fcoe_regex = re.compile("fcoe")
        if fc_regex.search(host_info["symbolic_name"]):
            return "FC"
        if fcoe_regex.search(host_info["symbolic_name"]):
            return "FCoE"

    if "driver" in list(host_info.keys()) and host_info["driver"] in ["bnx2fc", "qedf"]:
        return "FCoE"

    _print("DEBUG: Could not figure out if controller %s is FC or FCoE" % scsi_host_id)
    print(host_info)
    return "(TODO)FC/FCoE"


def get_value_rport_parameter(r_port, r_port_param):
    """Usage
        get_value_rport_parameter(r_port, r_port_param)
    Purpose
        Query out the r_port_param via sys:
        /sys/class/fc_remote_ports/rport-3:0-0/dev_loss_tmo
    Parameters
        r_port:         # like "rport0:0-1"
        r_port_param    # like "dev_loss_tmo"
    Returns
        parameter value
            or
        None.
    """
    if not r_port or not r_port_param:
        _print("FAIL: get_value_rport_parameter() - requires r_port and r_port_param as parameter")
        return None

    sys_r_port_param = f"{remote_port_path}/{r_port}/{r_port_param}"
    if not os.path.isfile(sys_r_port_param):
        _print("FAIL: get_value_rport_parameter() - File '%s' does not exist" % sys_r_port_param)
        return None

    ret, value = run("cat %s" % sys_r_port_param, return_output=True, verbose=False)
    if ret != 0:
        _print("FAIL: get_value_rport_parameter() - Could not read %s" % sys_r_port_param)
        # print command output
        print(value)
        return None

    return value


def set_value_rport_parameter(r_port, r_port_param, value):
    """Usage
        set_value_rport_parameter(r_port, r_port_param, value)
    Purpose
        Write value to r_port_param via sys:
        /sys/class/fc_remote_ports/rport-3:0-0/dev_loss_tmo
    Parameters
        r_port:         # like "rport-0:0-1"
        r_port_param    # like "dev_loss_tmo"
        value           # like "60"
    Returns
        True
            or
        False.
    """
    if not r_port or not r_port_param or value is None:
        _print("FAIL: set_value_rport_parameter() - requires r_port, r_port_param and value as parameter")
        return False

    sys_r_port_param = f"{remote_port_path}/{r_port}/{r_port_param}"
    if not os.path.isfile(sys_r_port_param):
        _print("FAIL: set_value_rport_parameter() - File '%s' does not exist" % sys_r_port_param)
        return False

    ret, output = run(f"echo {value} > {sys_r_port_param}", return_output=True, verbose=False)
    if ret != 0:
        _print(f"FAIL: set_value_rport_parameter() - Could not set {sys_r_port_param} to {value}")
        # print command output
        print(output)
        return False
    return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""md.py: Module to manipulate MD devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os.path
import re

from libsan import _print
from libsan.host.cmdline import exists, run


def _mdadm_query(md_device, verbose=False):
    if not exists("mdadm"):
        if verbose:
            _print("INFO: mdadm is not installed")
        return None

    cmd = "mdadm -D /dev/%s" % md_device
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: couldn't query %s" % md_device)
        if verbose:
            print(output)
        return None
    return output


def md_query(verbose=False):
    """Query Soft RAID devices.
    The arguments are:
    \tverbose: Print additional information. Default: False
    Returns:
    \tdict: Return a list of md devices.
    """
    mdstat_file = "/proc/mdstat"

    if not os.path.exists(mdstat_file):
        if verbose:
            _print("INFO: there is no MD device")
        return False

    cmd = "cat %s" % mdstat_file
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        if verbose:
            _print("INFO: there is no MD device")
        return None

    md_name_regex = r"^(md\d+) :"
    md_devices = []
    for line in output.split("\n"):
        m = re.match(md_name_regex, line)
        if not m:
            continue
        md_devices.append(m.group(1))

    return md_devices


def md_get_info(md_device, verbose=False):
    """Query information of an MD device.
    The arguments are:
    \tmd_device: md device name to get information about
    \tverbose: Print additional information. Default: False
    Returns:
    \tdict: Return a dictionary with details about the md device.
    """
    if not md_device:
        return None

    if md_device not in md_query():
        if verbose:
            _print("INFO: %s is not a MD device" % md_device)
        return None

    output = _mdadm_query(md_device, verbose)
    if not output:
        return None

    md_info = {}
    # Try to get general information about the device
    md_info_regex = r"\s+(.*) : (.*)"
    for line in output.split("\n"):
        info_match = re.match(md_info_regex, line)
        if not info_match:
            continue
        info_name = info_match.group(1).lower()
        info_name = info_name.replace(" ", "_")
        md_info[info_name] = info_match.group(2)

    # Try to get the storage devices linked to the MD
    storage_section = False
    storage_regex = r"\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(.*)\s+(\S+)$"
    for line in output.split("\n"):
        if re.search(r"Number\s+Major\s+Minor\s+RaidDevice\s+State", line):
            storage_section = True
            md_info["storage_devices"] = {}
        if not storage_section:
            continue
        storage_match = re.match(storage_regex, line)
        if not storage_match:
            continue
        storage_info = {}
        storage_info["number"] = storage_match.group(1)
        storage_info["major"] = storage_match.group(2)
        storage_info["minor"] = storage_match.group(3)
        storage_info["raid_device"] = storage_match.group(4)
        storage_info["state"] = storage_match.group(5).strip()
        md_info["storage_devices"][storage_match.group(6)] = storage_info

    return md_info


def md_get_storage_dev(md_device, verbose=False):
    """Get the storage devices of an MD device.
    The arguments are:
    \tmd_device: md device name to get information about
    \tverbose: Print additional information. Default: False
    Returns:
    \tlist: Return a list of storage devices.
    """
    if not md_device:
        return None

    md_info = md_get_info(md_device, verbose)
    if not md_info:
        return None

    if "storage_devices" not in md_info:
        return None
    return md_info["storage_devices"].keys()


def md_stop(md_device, verbose=False):
    """Stop an specific md device.
    The arguments are:
    \tmd_device: md device name to get information about
    \tverbose: Print additional information. Default: False
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t\tFalse in case of failure.
    """
    cmd = "mdadm --stop /dev/%s" % md_device
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: couldn't stop %s" % md_device)
        if verbose:
            print(output)
        return False
    return True


def md_clean(device, verbose=False):
    """clean an specific storage device.
    The arguments are:
    \tdevice: storage device like /dev/sda
    \tverbose: Print additional information. Default: False
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t\tFalse in case of failure.
    """
    cmd = "mdadm --zero-superblock %s" % device
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: couldn't clean %s" % device)
        if verbose:
            print(output)
        return False
    return True


def md_remove(md_device, clean=False, verbose=False):
    """Remove an specific md device.
    The arguments are:
    \tmd_device: md device name to get information about
    \tverbose: Print additional information. Default: False
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t\tFalse in case of failure.
    """
    sto_devices = md_get_storage_dev(md_device)

    if not md_stop(md_device):
        return False

    cmd = "mdadm --remove /dev/%s" % md_device
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if (
        retcode != 0
        and
        # error opening the device can be ignored
        "mdadm: error opening /dev/%s: No such file or directory" % md_device not in output
    ):
        _print("FAIL: couldn't remove %s" % md_device)
        if verbose:
            print(output)
        return False
    if clean and sto_devices:
        for device in sto_devices:
            if not md_clean(device, verbose):
                return False
    return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""nxos.py: Module to handle commands on Cisco NXOS Switch."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex
import time

import libsan.host.linux
import libsan.host.ssh
from libsan import _print
from libsan.host.net import standardize_mac


def _standardize_port_id(port_id):
    """ """
    if not port_id:
        return None
    m = re.match(r"e(\S+)$", port_id)
    if m:
        return "Ethernet%s" % m.group(1)

    m = re.match(r"Eth(\S+)$", port_id)
    if m:
        return "Ethernet%s" % m.group(1)

    m = re.match(r"p(\S+)$", port_id)
    if m:
        return "port-channel%s" % m.group(1)

    m = re.match(r"Po(\S+)$", port_id)
    if m:
        return "port-channel%s" % m.group(1)

    m = re.match(r"vlan(\S+)$", port_id)
    if m:
        return "Vlan%s" % m.group(1)

    return port_id


class nxos:  # noqaN801
    """Class to manage Cisco NXOS switches"""

    host = None
    user = None
    passwd = None

    def __init__(self, hostname, username, password):
        self.host = hostname
        self.user = username
        self.passwd = password

    def _run(self, cmd, return_output=False, verbose=False, invoke_shell=False, expect=None):
        session = libsan.host.ssh.connect(self.host, user=self.user, passwd=self.passwd)
        if not session:
            _print("FAIL: Could not connect to switch (%s)" % self.host)
            if return_output:
                return 127, None
            return 127

        _print("INFO: Connecting to %s for command:" % self.host)
        print(cmd)
        ret, output = libsan.host.ssh.run_cmd(
            session, cmd, return_output=True, verbose=verbose, invoke_shell=invoke_shell, expect=expect
        )
        libsan.host.ssh.disconnect(session)
        if return_output:
            return ret, output
        return ret

    @staticmethod
    def capability():
        """Indicates supported operation on switch"""
        cap_dict = {}
        cap_dict["link_down"] = True
        cap_dict["link_up"] = True
        cap_dict["link_oscillate"] = True
        cap_dict["switch_reboot"] = False
        cap_dict["fc_switch"] = True
        cap_dict["eth_switch"] = True
        return cap_dict

    def get_version(self):
        """Return FW version of the switch"""
        cmd = "show version"
        _, output = self._run(cmd, return_output=True, verbose=False)
        nxos_regex = re.compile(r"Cisco Nexus Operating System", re.MULTILINE)
        version_regex = re.compile(r".*system:\s+version\s+(\d+)(\.\S+)", re.MULTILINE)
        n = nxos_regex.search(output)
        v = version_regex.search(output)
        if n and v:
            return f"{v.group(1)}{v.group(2)}"
        return None

    def show_interfaces(self, option="brief"):
        cmd = "show interface %s" % option
        return self._run(cmd)

    def link_trigger(self, action=None, port_id=None, check_state=True):
        if not action or not port_id:
            _print("FAIL: link_trigger needs 'action' and 'port_id' parameters")
            return False

        port_id = _standardize_port_id(port_id)
        if not port_id:
            _print("FAIL: link_trigger() - Invalid port format")
            return False

        cmds = ["terminal length 0", "configure terminal", "interface %s" % port_id]

        if action == "UP":
            cmds.append("no shutdown")
        elif action == "DOWN":
            cmds.append("shutdown")
        else:
            _print("FAIL: link_trigger unsupported action (%s)" % action)
            return False
        # As the command to change interface is inside configure terminal, we need to run all commands in a single
        # shell, so we need to use invoke_shell
        ret = self._run(cmds, verbose=False, invoke_shell=True, expect="# ")
        if ret != 0:
            # It will never hit this point as _run with invoke_shell always return success
            _print("FAIL: link_trigger command failed")
            return False

        # We do not want make sure the state is correct
        if not check_state:
            return True

        attempts = 0
        max_attempts = 5
        verbose = False
        # Try few times to read port state
        while attempts < max_attempts:
            attempts += 1
            # wait 5s to detect link change
            time.sleep(5)
            p_state_dict = self.query_port_state(verbose=verbose)
            if not p_state_dict:
                _print("FATAL: Could not read switch port state")
                continue

            if port_id not in list(p_state_dict.keys()):
                _print("FAIL: Could not find state for port %s" % port_id)
                print(p_state_dict)
                # If could not get the state correct in the first time,
                # print query state output from the switch
                verbose = True
                continue

            if p_state_dict[port_id] == action:
                _print(f"INFO: Successfully brought port {port_id} {action}")
                return True

            if attempts == max_attempts:
                _print(f"FAIL: Failed to bring port {port_id} {action}. Current state: {p_state_dict[port_id]}")
                return False
        return False

    def port_state(self, port_id):
        port_id = _standardize_port_id(port_id)
        if not port_id:
            _print("FAIL: port_state requires 'port_id' as parameter")
            return None

        p_state_dict = self.query_port_state()
        if not p_state_dict:
            _print("FAIL: could not query port_state on switch")
            return None

        if port_id not in list(p_state_dict.keys()):
            _print("FAIL: port_state %s was not found on switch" % port_id)
            return None

        return p_state_dict[port_id]

    def port_connect(self, port_id, check_state=True):
        return self.link_trigger(action="UP", port_id=port_id, check_state=check_state)

    def port_disconnect(self, port_id, check_state=True):
        return self.link_trigger(action="DOWN", port_id=port_id, check_state=check_state)

    def port_oscillate(self, port_id, min_uptime, max_uptime, min_downtime, max_downtime, count):
        if (
            not port_id
            or min_uptime is None
            or max_uptime is None
            or min_downtime is None
            or max_downtime is None
            or count is None
        ):
            _print(
                "FAIL: port_oscillate() - needs 'port_id' , 'min_uptime', 'max_uptime', "
                "'min_downtime', 'max_downtime' and 'count' parameters"
            )
            return False

        for cnt in range(1, count + 1):
            _print("INFO: Interaction %d/%d" % (cnt, count))
            self.port_disconnect(port_id, check_state=False)
            increment = (max_downtime - min_downtime) / count
            sleep_time = min_downtime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link down" % sleep_time)
            libsan.host.linux.sleep(sleep_time)
            self.port_connect(port_id, check_state=False)
            increment = (max_uptime - min_uptime) / count
            sleep_time = min_uptime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link up" % sleep_time)
            libsan.host.linux.sleep(sleep_time)

        return True

    def query_port_state(self, verbose=False):
        """Query all interfaces on switch and store there status on a dict
        If port is trunking or 'up' we set it as UP
        Otherwise we set it as DOWN
        """
        cmd = "show interface"
        ret, output = self._run(cmd, return_output=True, verbose=verbose)
        if ret != 0:
            return None

        state_regex = re.compile(r"^(\S+) is ([a-z]+)")
        lines = output.split("\n")
        port_state = {}
        for line in lines:
            m = state_regex.match(line)
            if m:
                state = m.group(2)
                if state == "trunking" or state == "up":
                    port_state[m.group(1)] = "UP"
                else:
                    port_state[m.group(1)] = "DOWN"
        return port_state

    def wwpn_2_port_id(self):
        """Query all interfaces that are connected
        And store as key the WWPN and the Port ID as its value
        """
        cmd = "show flogi database"
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            return None

        """# show flogi database
--------------------------------------------------------------------------------
INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
--------------------------------------------------------------------------------
vfc1             702   0xb30000  20:00:90:e2:ba:a3:97:cb 10:00:90:e2:ba:a3:97:cb
        """
        parse_regex = re.compile(r"^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)")
        lines = output.split("\n")
        wwn_dict = {}
        for line in lines:
            m = parse_regex.match(line)
            if m:
                if m.group(1) == "Total" or m.group(1) == "INTERFACE":
                    continue
                wwpn = m.group(4)
                wwn_dict[m.group(1)] = wwpn
        return wwn_dict

    def mac_2_port_id(self):
        """Query all interfaces that are connected
        And store as key the WWPN and the Port ID as its value
        """
        cmd = "show mac address-table"
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: mac_2_port_id")
            print(output)
            return None

        """# show mac address-table
        Legend:
                * - primary entry, G - Gateway MAC, (R) - Routed MAC, O - Overlay MAC
                age - seconds since last seen,+ - primary entry using vPC Peer-Link
        VLAN     MAC Address      Type      age     Secure NTFY   Ports/SWID.SSID.LID
        ---------+-----------------+--------+---------+------+----+------------------
        * 1        90e2.baa3.9528    dynamic   20         F    F  Eth1/10
        """
        parse_regex = re.compile(r"^\*\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)")
        lines = output.split("\n")
        mac_dict = {}
        for line in lines:
            m = parse_regex.match(line)
            if m:
                mac = standardize_mac(m.group(2))
                port_id = _standardize_port_id(m.group(7))
                if not port_id:
                    _print("FAIL: mac_2_port_id() - Could not understand port_id %s format" % m.group(7))
                    return None
                mac_dict[port_id] = mac
        return mac_dict



"""__init__.py: general methods used on libsan"""
import inspect
import re
import sys


def _print(string):
    # name of the module that called the function
    module_name = inspect.currentframe().f_back.f_globals["__name__"]
    string = re.sub("DEBUG:", "DEBUG:(" + module_name + ")", string)
    string = re.sub("FAIL:", "FAIL:(" + module_name + ")", string)
    string = re.sub("FATAL:", "FATAL:(" + module_name + ")", string)
    string = re.sub("WARN:", "WARN:(" + module_name + ")", string)
    print(string)
    sys.stdout.flush()
    if "FATAL:" in string:
        raise RuntimeError(string)



#!/usr/bin/python

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

import os

import pytest
from libsan.host import dt, linux


def test_dt():
    if not dt.install_dt():
        pytest.fail("Unable to install dt")

    if linux.is_docker():
        print("SKIP: DT has a bug and does not work on container")
        return

    output_file = "dt.test"
    try:
        dt_pid = dt.dt_stress_background(output_file, limit="100m", verbose=True)
    except Exception as e:
        pytest.fail("FAIL: Exception: %s" % e)

    # Make sure background process does not generate exception
    print("INFO: Waiting DT process to finish")
    try:
        _, exit_status = os.waitpid(dt_pid, 0)
    except Exception as e:
        pytest.fail("FAIL: Exception: %s" % e)

    if exit_status != 0:
        pytest.fail("FAIL: there was some error running DT")

    if not os.remove(output_file):
        pytest.fail("FAIL: Could not delete %s" % output_file)

    assert 1
